"""
written in python from 
from original TU Vienna codes for GMF
"""
import datetime
import math
import os
import pickle
import subprocess
import sys
import wget

import numpy as np
from scipy.interpolate import interp1d

import gnssrefl.gps as g


def read_4by5(station, dlat,dlon,hell):
    """
    reads existing grid points for a given location

    Parameters
    ----------
    station : string
        name of station
    dlat : float
        latitude in degrees
    dlon : float
        longitude in degrees
    hell : float
        ellipsoidal height in meters

    Returns
    -------
    pgrid : 4 by 5 numpy array
        pressure in hPa

    Tgrid : 4 by 5 numpy array 
        temperature in C

    Qgrid : 4 by 5 numpy array 

    dTgrid : 4 by 5 numpy array  
       temperature lapse rate in degrees per km 

    u : 4 by 1 numpy array
        geoid undulation in meters

    Hs : 4 by 1 numpy array  

    ahgrid : 4 by 5 numpy array
        hydrostatic mapping function coefficient at zero height (VMF1) 

    awgrid : 4 by 5 numpy array
        wet mapping function coefficient (VMF1) 

    lagrid : 4 by 5 numpy array

    Tmgrid : 4 by 5 numpy array
        mean temperature of the water vapor in degrees Kelvin 

    requires that an environment variable exists for REFL_CODE
    """
#
    xdir = str(os.environ['REFL_CODE'])
    inputpath = xdir + '/input/'
#    if not os.path.isdir(inputpath): #if year folder doesn't exist, make it
#        os.makedirs(inputpath)

    # input file should be written here
    obsfile = inputpath + station + '_refr.txt'
    #print('reading from station refraction file: ', obsfile)
    x = np.genfromtxt(obsfile,comments='%')
    max_ind = 4
    pgrid = np.zeros((4,5))
    Tgrid = np.zeros((4,5))
    Qgrid = np.zeros((4,5))
    dTgrid = np.zeros((4,5))
    u = np.zeros((4,1))
    Hs = np.zeros((4,1)) 
    ahgrid = np.zeros((4,5))
    awgrid = np.zeros((4,5))
    lagrid = np.zeros((4,5))
    Tmgrid = np.zeros((4,5))

    for n in [0,1,2,3]:
        ij = 0
        u[n]= x[n*5,6]
        Hs[n]= x[n*5,7]
        for m in range(n*5, n*5+5):
            pgrid[n,ij] = x[m,2] 
            Tgrid[n,ij] = x[m,3] 
            Qgrid[n,ij] = x[m,4]/1000
            dTgrid[n,ij] = x[m,5]/1000
            ahgrid[n,ij] = x[m,8]/1000
            awgrid[n,ij] = x[m,9]/1000
            lagrid[n,ij] = x[m,10] 
            Tmgrid[n,ij] = x[m,11] 
            ij +=1

    return pgrid, Tgrid, Qgrid, dTgrid, u, Hs, ahgrid, awgrid, lagrid, Tmgrid
#
def gpt2_1w (station, dmjd,dlat,dlon,hell,it):
    """
    Parameters
    ----------
    station : str
        station name
    dmjd:  float 
        modified Julian date (scalar, only one epoch per call is possible)
    dlat : float 
        ellipsoidal latitude in radians [-pi/2:+pi/2] 
    dlon : float
        longitude in radians [-pi:pi] or [0:2pi] 
    hell : float 
        ellipsoidal height in m 
    it: integer
        case 1: no time variation but static quantities

        case 0: with time variation (annual and semiannual terms)

    Returns
    -------
    p : float
        pressure in hPa
    T:  float
        temperature in degrees Celsius 
    dT : float
       temperature lapse rate in degrees per km 
    Tm : float
        mean temperature of the water vapor in degrees Kelvin 
    e : float
        water vapor pressure in hPa 
    ah: float
        hydrostatic mapping function coefficient at zero height (VMF1) 
    aw: float
        wet mapping function coefficient (VMF1) 
    la: float
        water vapor decrease factor 
    undu: float
        geoid undulation in m 

    """

#  need to find diffpod and difflon
    if (dlon < 0):
        plon = (dlon + 2*np.pi)*180/np.pi;
    else:
        plon = dlon*180/np.pi;
# transform to polar distance in degrees
    ppod = (-dlat + np.pi/2)*180/np.pi; 

#       % find the index (line in the grid file) of the nearest point
#  	  % changed for the 1 degree grid (GP)
    ipod = np.floor(ppod+1); 
    ilon = np.floor(plon+1);
    
#   normalized (to one) differences, can be positive or negative
#	% changed for the 1 degree grid (GP)
    diffpod = (ppod - (ipod - 0.5));
    difflon = (plon - (ilon - 0.5));


# change the reference epoch to January 1 2000
    #print('Modified Julian Day', dmjd)
    dmjd1 = dmjd-51544.5 

    pi2 = 2*np.pi
    pi4 = 4*np.pi

# mean gravity in m/s**2
    gm = 9.80665;
# molar mass of dry air in kg/mol
    dMtr = 28.965E-3 
#    dMtr = 28.965*10^-3 
# universal gas constant in J/K/mol
    Rg = 8.3143 

# factors for amplitudes, i.e. whether you want time varying
    if (it==1):
        #print('>>>> no refraction time variation ')
        cosfy = 0; coshy = 0; sinfy = 0; sinhy = 0;
    else: 
        cosfy = np.cos(pi2*dmjd1/365.25)
        coshy = np.cos(pi4*dmjd1/365.25) 
        sinfy = np.sin(pi2*dmjd1/365.25) 
        sinhy = np.sin(pi4*dmjd1/365.25) 
    cossin = np.matrix([1, cosfy, sinfy, coshy, sinhy])
# initialization of new vectors
    p =  0; T =  0; dT = 0; Tm = 0; e =  0; ah = 0; aw = 0; la = 0; undu = 0;
    undul = np.zeros(4)
    Ql = np.zeros(4)
    dTl = np.zeros(4)
    Tl = np.zeros(4)
    pl = np.zeros(4)
    ahl = np.zeros(4)
    awl = np.zeros(4)
    lal = np.zeros(4)
    Tml = np.zeros(4)
    el = np.zeros(4)
#
    pgrid, Tgrid, Qgrid, dTgrid, u, Hs, ahgrid, awgrid, lagrid, Tmgrid = read_4by5(station,dlat,dlon,hell)
#
    for l in [0,1,2,3]:
        KL = l   #silly to have this as a variable like this 
#  transforming ellipsoidal height to orthometric height:
#  Hortho = -N + Hell
        undul[l] = u[KL] 
        hgt = hell-undul[l] 
#  pressure, temperature at the height of the grid
        T0 = Tgrid[KL,0] + Tgrid[KL,1]*cosfy + Tgrid[KL,2]*sinfy + Tgrid[KL,3]*coshy + Tgrid[KL,4]*sinhy;
        tg = float(Tgrid[KL,:] *cossin.T)
#     print(T0,tg)

        p0 = pgrid[KL,0] + pgrid[KL,1]*cosfy + pgrid[KL,2]*sinfy + pgrid[KL,3]*coshy + pgrid[KL,4]*sinhy;
 
#       humidity 
        Ql[l] = Qgrid[KL,0] + Qgrid[KL,1]*cosfy + Qgrid[KL,2]*sinfy + Qgrid[KL,3]*coshy + Qgrid[KL,4]*sinhy;
 
# reduction = stationheight - gridheight
        Hs1 = Hs[KL]
        redh = hgt - Hs1;

# lapse rate of the temperature in degree / m
        dTl[l] = dTgrid[KL,0] + dTgrid[KL,1]*cosfy + dTgrid[KL,2]*sinfy + dTgrid[KL,3]*coshy + dTgrid[KL,4]*sinhy;
   
# temperature reduction to station height
        Tl[l] = T0 + dTl[l]*redh - 273.15;

#  virtual temperature
        Tv = T0*(1+0.6077*Ql[l])   
        c = gm*dMtr/(Rg*Tv) 
        
# pressure in hPa
        pl[l] = (p0*np.exp(-c*redh))/100 
            
#  hydrostatic coefficient ah
        ahl[l] = ahgrid[KL,0] + ahgrid[KL,1]*cosfy + ahgrid[KL,2]*sinfy + ahgrid[KL,3]*coshy + ahgrid[KL,4]*sinhy;
            
# wet coefficient aw
        awl[l] = awgrid[KL,0] + awgrid[KL,1]*cosfy + awgrid[KL,2]*sinfy + awgrid[KL,3]*coshy + awgrid[KL,4]*sinhy;
					 
# water vapor decrease factor la - added by GP
        lal[l] = lagrid[KL,0] + lagrid[KL,1]*cosfy + lagrid[KL,2]*sinfy + lagrid[KL,3]*coshy + lagrid[KL,4]*sinhy;
					 
# mean temperature of the water vapor Tm - added by GP
        Tml[l] = Tmgrid[KL,0] +  Tmgrid[KL,1]*cosfy + Tmgrid[KL,2]*sinfy + Tmgrid[KL,3]*coshy + Tmgrid[KL,4]*sinhy;
					 		 
# water vapor pressure in hPa - changed by GP
        e0 = Ql[l]*p0/(0.622+0.378*Ql[l])/100; # % on the grid
        aa = (100*pl[l]/p0)
        bb = lal[l]+1
        el[l] = e0*np.power(aa,bb)  # % on the station height - (14) Askne and Nordius, 1987
           
    dnpod1 = np.abs(diffpod); # % distance nearer point
    dnpod2 = 1 - dnpod1;   # % distance to distant point
    dnlon1 = np.abs(difflon);
    dnlon2 = 1 - dnlon1;
        
#   pressure
    R1 = dnpod2*pl[0]+dnpod1*pl[1];
    R2 = dnpod2*pl[2]+dnpod1*pl[3];
    p = dnlon2*R1+dnlon1*R2;
            
#   temperature
    R1 = dnpod2*Tl[0]+dnpod1*Tl[1];
    R2 = dnpod2*Tl[2]+dnpod1*Tl[3];
    T = dnlon2*R1+dnlon1*R2;
        
#   temperature in degree per km
    R1 = dnpod2*dTl[0]+dnpod1*dTl[1];
    R2 = dnpod2*dTl[2]+dnpod1*dTl[3];
    dT = (dnlon2*R1+dnlon1*R2)*1000;
            
#   water vapor pressure in hPa - changed by GP
    R1 = dnpod2*el[0]+dnpod1*el[1];
    R2 = dnpod2*el[2]+dnpod1*el[3];
    e = dnlon2*R1+dnlon1*R2;
            
#   hydrostatic
    R1 = dnpod2*ahl[0]+dnpod1*ahl[1];
    R2 = dnpod2*ahl[2]+dnpod1*ahl[3];
    ah = dnlon2*R1+dnlon1*R2;
           
#   wet
    R1 = dnpod2*awl[0]+dnpod1*awl[1];
    R2 = dnpod2*awl[2]+dnpod1*awl[3];
    aw = dnlon2*R1+dnlon1*R2;
        
#  undulation
    R1 = dnpod2*undul[0]+dnpod1*undul[1];
    R2 = dnpod2*undul[2]+dnpod1*undul[3];
    undu = dnlon2*R1+dnlon1*R2;

#   water vapor decrease factor la - added by GP
    R1 = dnpod2*lal[0]+dnpod1*lal[1];
    R2 = dnpod2*lal[2]+dnpod1*lal[3];
    la = dnlon2*R1+dnlon1*R2;
		
#   mean temperature of the water vapor Tm - added by GP
    R1 = dnpod2*Tml[0]+dnpod1*Tml[1];
    R2 = dnpod2*Tml[2]+dnpod1*Tml[3];
    Tm = dnlon2*R1+dnlon1*R2; 

    return p, T, dT,Tm,e,ah,aw,la,undu

def readWrite_gpt2_1w(xdir, station, site_lat, site_lon):
    """
    makes a grid for refraction correction

    Parameters
    ----------
    xdir : str
        directory for output
    station : str
        station name, 4 ch
    lat : float
        latitude in degrees
    lon : float 
        longitude in degrees 

    """
    PROJECT_ROOT = os.path.dirname(os.path.abspath(__file__))
    BASE_DIR = os.path.dirname(PROJECT_ROOT)
    try3 = PROJECT_ROOT + '/' + 'gpt_1wA.pickle'

    # check that output path exists.  
    outpath = xdir + '/input/'
    if not os.path.isdir(outpath): 
        os.makedirs(outpath)

    outfile = outpath + station + '_refr.txt'
    if os.path.isfile(outfile):
        okokok = 1
        #print('refraction file for this station already exists')
    else:
        print('A station specific refraction output file will be written to ', outfile)

#   change to radians
        dlat = site_lat*np.pi/180 
        dlon = site_lon*np.pi/180 

#   read VMF gridfile in pickle format 
        pname = xdir + '/input/' + 'gpt_1wA.pickle'
        foundit, pname = look_for_pickle_file() 
        if foundit:
            f = open(pname, 'rb')
            [All_pgrid, All_Tgrid, All_Qgrid, All_dTgrid, All_U, All_Hs, All_ahgrid, All_awgrid, All_lagrid, All_Tmgrid] = pickle.load(f)
            f.close()
        else:
            print('You will need to download gpt_1wA.pickle MANUALLY from github and store it in REFL_CODE/input')
            sys.exit()

# really should e zero to four, but whatever
        indx = np.zeros(4,dtype=int)
        indx_lat = np.zeros(4,dtype=int)
        indx_lon = np.zeros(4,dtype=int)


#figure out grid index
# % only positive longitude in degrees
        if (dlon < 0):
            plon = (dlon + 2*np.pi)*180/np.pi;
        else:
            plon = dlon*180/np.pi 
#
#  transform to polar distance in degrees
        ppod = (-dlat + np.pi/2)*180/np.pi  

#% find the index (line in the grid file) of the nearest point
# % changed for the 1 degree grid (GP)
        ipod = np.floor(ppod+1)  
        ilon = np.floor(plon+1) 
    
#    % normalized (to one) differences, can be positive or negative
# % changed for the 1 degree grid (GP)
        diffpod = (ppod - (ipod - 0.5)) 
        difflon = (plon - (ilon - 0.5)) 
#    % added by HCY
# % changed for the 1 degree grid (GP)
        if (ipod == 181):
            ipod = 180 
        if (ilon == 361):
            ilon = 1 
        if (ilon == 0):
            ilon = 360

#     get the number of the corresponding line
#	 changed for the 1 degree grid (GP)
        indx[0] = (ipod - 1)*360 + ilon 
#  save the lat lon of the grid points
        indx_lat[0] = 90-ipod+1  
        indx_lon[0] = ilon-1   
# % near the poles: nearest neighbour interpolation, otherwise: bilinear
# % with the 1 degree grid the limits are lower and upper (GP)

        bilinear = 0 
        max_ind = 1 
        if (ppod > 0.5) and (ppod < 179.5):
            bilinear = 1           
        if (bilinear == 1):
            max_ind =4 

#    % bilinear interpolation
#    % get the other indexes 
 
        ipod1 = ipod + np.sign(diffpod) 
        ilon1 = ilon + np.sign(difflon) 
# % changed for the 1 degree grid (GP)
        if (ilon1 == 361):
            ilon1 = 1 
        if (ilon1 == 0):
            ilon1 = 360 
#         get the number of the line
# changed for the 1 degree grid (GP)
# four indices ???
        indx[1] = (ipod1 - 1)*360 + ilon; # % along same longitude
        indx[2] = (ipod  - 1)*360 + ilon1;# % along same polar distance
        indx[3] = (ipod1 - 1)*360 + ilon1;# % diagonal
#
# save the lat lon of the grid points  lat between [-90 ;90]  lon [0 360] 
        indx_lat[1] =   90 - ipod1+np.sign(diffpod)     
        indx_lon[1] = ilon-1 
        indx_lat[2] =   90-ipod +1
        indx_lon[2] =  ilon1 - np.sign(difflon) 
        indx_lat[3] =   90 -ipod1+np.sign(diffpod)     
        indx_lon[3] = ilon1- np.sign(difflon);

# extract the new grid
# will need to do 0-4 instead of 1-5 because stored that way in python
# which values to use in the bigger array
# assign the correct values
        indx = indx - 1
        indx_list = indx.tolist()
#    print(indx_list)
#    print(indx)
#print(np.shape(indx_lat))
#print(np.shape(indx_lon))
        w = 0
# need to write values for a given station to a plain text file
#
        fout = open(outfile, 'w+')
        for a in indx_list:
            for k in [0,1,2,3,4]:
                fout.write(" {0:4.0f} {1:5.0f} {2:13.4f} {3:10.4f} {4:10.6f} {5:10.4f} {6:12.5f} {7:12.5f} {8:10.6f} {9:10.6f} {10:10.6f} {11:10.4f} \n".format( indx_lat[w], indx_lon[w],All_pgrid[a,k],All_Tgrid[a,k],All_Qgrid[a,k]*1000,All_dTgrid[a,k]*1000,All_U[a,0],All_Hs[a,0], All_ahgrid[a,k]*1000, All_awgrid[a,k]*1000, All_lagrid[a,k], All_Tmgrid[a,k] ))

            w+=1
        fout.close()
        print('station specific refraction file written')


def corr_el_angles(el_deg, press, temp):
    """
    Corrects elevation angles for refraction using simple angle bending model

    Parameters
    ----------
    el_deg : numpy array of floats
        elevation angles in degrees
    press : float
        pressure in hPa
    temp : float
        temperature in degrees C

    Returns
    -------
    corr_el_deg : numpy array of floats
        corrected elevation angles (in degrees)

    """

#  Formula in python from Strandberg, originally from Astronomy journal
    corr_el_arc_min = 510/(9/5*temp + 492) * press/1010.16 * 1/np.tan(np.deg2rad(el_deg + 7.31/(el_deg + 4.4)))
    correction = corr_el_arc_min/60 
     
    corr_el_deg = el_deg + correction   
    return corr_el_deg


def look_for_pickle_file():
    """
    latest attempt to solve the dilemma of the pickle file needed for
    the refraction correction

    Returns
    -------
    foundit : bool
        whether pickle file found
    fullpname : str
        full path to the pickle file
    """
#   read VMF gridfile in pickle format

    pfile = 'gpt_1wA.pickle'

    # at one point i thought this was useful
    PROJECT_ROOT = os.path.dirname(os.path.abspath(__file__))
    BASE_DIR = os.path.dirname(PROJECT_ROOT)
    try3 = PROJECT_ROOT + '/' + pfile
    # do i need str??
    xdir = str(os.environ['REFL_CODE'])
    inputdir = xdir + '/input/'
    if os.path.isdir(inputdir):
        print('found ', inputdir)
    else:
        print('make directory: ', inputdir)
        subprocess.call(['mkdir',inputdir])

    fdir = xdir + '/Files'
    if not os.path.isdir(fdir):
        print('make Files directory: ', fdir)
        subprocess.call(['mkdir', fdir])

    # where the file should be stored
    fullpname = inputdir + pfile

    print('The large refraction file should be stored here:', fullpname)
    foundit = False

    if os.path.isfile(fullpname):
        print('1st attempt: found in ', fullpname)
        foundit = True
    else:
        print('1st attempt: not found in ', fullpname)

    if not foundit:
        pname =  'gnssrefl/gpt_1wA.pickle'
        print('2nd attempt: look in subdirectory of current working directory:', pname)
        if os.path.isfile(pname):
            subprocess.call(['cp','-f',pname, fullpname  ])
            foundit = True

    if not foundit:
        pname = try3
        print('3rd attempt try looking here: ',pname)
        if os.path.isfile(pname):
            foundit = True
            print('cp it to ', fullpname)
            subprocess.call(['cp','-f',pname, fullpname])
        else:
            print('that did not work')

    url= 'https://github.com/kristinemlarson/gnssrefl/raw/master/gnssrefl/' + pfile
    if not foundit:
        print('4th attempt - download from github', url)
        try:
            wget.download(url,fullpname)
        except:
            print('download or cp failed')
        if os.path.isfile(fullpname):
            foundit = True

    if not foundit:
        url= 'https://morefunwithgps.com/public_html/' + pfile
        print('5th attempt - ',url)
        try:
            wget.download(url,fullpname)
        except:
            print('Failed again.')

        if os.path.isfile(fullpname):
            foundit = True
        else:
            print('File should be stored in ', inputdir, ' but is not')

    return foundit , fullpname


def Ulich_Bending_Angle(ele, N0,lsp,p,T,ttime,sat):   #UBA
    """
    Ulich, B. L. "Millimeter wave radio telescopes: Gain and pointing characteristics." (1981)

    Author: 20220629, fengpeng

    Modified by KL to use numpy so I can use arrays.  I do not know why
    all these extra input parameters are here.

    Parameters
    ----------
    ele : numpy array of floats
        true elevation angle, degrees

    N0 : float
        antenna refractivity in ppm

    lsp : dict

    p : float
        pressure, units?

    T :  float
        temperature, units?

    ttime : 

    sat : 

    Returns
    -------
    De : numpy array of floats
        corrected elevation angles, deg
        
    """
    e_simple_corr = corr_el_angles(ele, p,T)
    deg2rad = np.pi/180

    # change to radians
    ele_rad = deg2rad*(ele)
    r = N0/1000000.
    f = np.cos(ele_rad) / (np.sin(ele_rad) + 0.00175 * np.tan(deg2rad*(87.5) - ele_rad))
    dE = (r * f)/deg2rad
    #fout = open('ulich.txt', 'w+')
    #for i in range(0,len(ele)):
    #    fout.write(" {0:8.5f} {1:8.5f} {2:8.5f} {3:10.0f} {4:5.0f} \n ".format( 
    #        ele[i], ele[i]+dE[i], e_simple_corr[i], ttime[i], sat[i]))
    #fout.close()

    return dE + ele

def refrc_Rueger(drypress,vpress,temp):
    """
    Obtains refractivity index suitable for GNSS-IR

    Rueger, Jean M. "Refractive index formulae for radio waves." Proceedings of the 
    FIG XXII International Congress, Washington, DC, USA. Vol. 113. 2002.

    Parameters
    ----------
    drypress : float
        dry pressure hPa
    vpress : float
        vapor pressure in hPa
    temp : float
        temperature in Kelvin

    Returns 
    -------
    ref : list of floats
         [Ntotal, Nhydro, Nwet], which are total, hydrostatic and wet refractivity in ppm

    """

    # Rueger's "best average", for 375 ppm CO2, 77.690 for 392 ppm CO2
    [K1r,K2r,K3r]=[77.689 ,71.2952 ,375463.]     

    # Rueger,IAG recommend, in ppm
    Nrueger = K1r * drypress / temp + K2r * vpress / temp + K3r * vpress / (temp ** 2)  

    # gas constant (SI exact), molar masses of dry air and water, kg/mol
    [Rgas,Md,Mw] = [8.31446261815324,0.0289644,0.01801528]   

    #density in kg/m^3
    drydensity=(drypress*100.)*Md/(Rgas*temp)    
    #in PV=nRT, P in Pa, V in m^3, n in g/mol, T in K, R=8.314 Pa*m^3/(k*mol)
    vdensity=(vpress*100.)*Mw/(Rgas*temp)   
    totaldensity=drydensity+vdensity

    #divide by 100 because hPa in Rueger formula, result in ppm
    Nhydro=(K1r*Rgas*totaldensity/Md)/100.   
    K2rr=K2r-K1r*(Mw/Md)
    Nwet=K2rr * vpress / temp + K3r * vpress / (temp ** 2)
    # in ppm
    ref=[Nrueger,Nhydro,Nwet]    
    #ref=[round(Nrueger,4),round(Nhydro,4),round(Nwet,4)]    

    return ref


def Equivalent_Angle_Corr_NITE(Hr_apr, e_T, N_ant, ztd_ant, mpf_tot, dmpf_de_tot):
    """

    This function computes the "equvilent" angular correction to apply the 
    NITE formula on the true elevation angle ele_eqv = e_T + dele


    Equation (24) in Peng (2023), DOI: 10.1109/TGRS.2023.3332422

    The variable substitude method can be found in Strandberg, J. (2020). 
    New methods and applications for interferometric GNSS reflectometry. 
    Chalmers Tekniska Hogskola (Sweden).

    Parameters
    ----------
    Hr_apr : float
        approximate a-priori reflector height, in meters
    e_T : float
        satellite true elevation angle in degree
    N_ant : float
        atmospheric refractivity at the GNSS antenna, in ppm
    ztd_ant : float
        zenith total delay at the antenna, in meters
    mpf_tot : float
        total mapping function for this elevation angle
    dmpf_de_tot : float
        derivative of the mapping function over elevation angle

    Returns
    -------
    dele : float
        equvilent angular correction in degrees

    """
    e_A = e_T+Ulich_Bending_Angle_original(e_T, N_ant)
    sita_E = sita_Earth(Hr_apr, e_A)
    sita_S = sita_Satellite(Hr_apr, e_A)
    e_A_r = e_A+sita_E +sita_S
    Hv_ratio = Hv_Hr_ratio(Hr_apr, 6378137., e_A)
    sin_eqv_geo = 0.5 * Hv_ratio / np.sin(np.radians(e_A_r + sita_E)) * (1. - np.cos(np.radians(e_T + e_A_r + sita_E)))
    Nl = N_layer(N_ant, Hr_apr)
    dsin_up = Hv_ratio*(Nl / 1000000.)/np.sin(np.radians(e_A_r + sita_E))
    ddele_T_dh = np.tan(np.radians(e_A))
    dsin_down = Nl / 1000000.*mpf_tot-ztd_ant*(dmpf_de_tot*(1. / (6378137. *ddele_T_dh ))+dmpf_dh(e_T, dhgt=1.))
    dsin = 0.5*(dsin_up + dsin_down)
    ele_eqv = np.degrees(np.arcsin(sin_eqv_geo + dsin))
    dele = ele_eqv - e_T

    return dele


def gmf_deriv(dmjd,dlat,dlon,dhgt,zd):
    """
    This subroutine determines the Global Mapping Functions GMF  and derivative.
    Translated from https://vmf.geo.tuwien.ac.at/codes/gmf_deriv.f by Peng Feng in March, 2023.

    Johannes Boehm, 2005 August 30
    
    ref 2006 Aug. 14: derivatives (U. Hugentobler)
    ref 2006 Aug. 14: recursions for Legendre polynomials (O. Montenbruck)
    ref 2011 Jul. 21: latitude -> ellipsoidal latitude

    Parameters
    ----------
    dmjd: float
        modified julian date
    dlat: float
        ellipsoidal latitude in radians
    dlon: float
        longitude in radians
    dhgt: float
        height in meters
    zd: float
        zenith distance in radians ??? ( is this really what you mean??
        KL: I suspect it is the zenith angle ... in radians

    Returns
    -------
    gmfh(2): float
        hydrostatic mapping function and derivative wrt z

    gmfw(2): float
        wet mapping function and derivative wrt z
    
    """

    ah_mean=[+1.2517e+02, +8.503e-01, +6.936e-02, -6.760e+00, +1.771e-01,
             +1.130e-02, +5.963e-01, +1.808e-02, +2.801e-03, -1.414e-03,
             -1.212e+00, +9.300e-02, +3.683e-03, +1.095e-03, +4.671e-05,
             +3.959e-01, -3.867e-02, +5.413e-03, -5.289e-04, +3.229e-04,
             +2.067e-05, +3.000e-01, +2.031e-02, +5.900e-03, +4.573e-04,
             -7.619e-05, +2.327e-06, +3.845e-06, +1.182e-01, +1.158e-02,
             +5.445e-03, +6.219e-05, +4.204e-06, -2.093e-06, +1.540e-07,
             -4.280e-08, -4.751e-01, -3.490e-02, +1.758e-03, +4.019e-04,
             -2.799e-06, -1.287e-06, +5.468e-07, +7.580e-08, -6.300e-09,
             -1.160e-01, +8.301e-03, +8.771e-04, +9.955e-05, -1.718e-06,
             -2.012e-06, +1.170e-08, +1.790e-08, -1.300e-09, +1.000e-10]

    bh_mean=[+0.000e+00, +0.000e+00, +3.249e-02, +0.000e+00, +3.324e-02,
             +1.850e-02, +0.000e+00, -1.115e-01, +2.519e-02, +4.923e-03,
             +0.000e+00, +2.737e-02, +1.595e-02, -7.332e-04, +1.933e-04,
             +0.000e+00, -4.796e-02, +6.381e-03, -1.599e-04, -3.685e-04,
             +1.815e-05, +0.000e+00, +7.033e-02, +2.426e-03, -1.111e-03,
             -1.357e-04, -7.828e-06, +2.547e-06, +0.000e+00, +5.779e-03,
             +3.133e-03, -5.312e-04, -2.028e-05, +2.323e-07, -9.100e-08,
             -1.650e-08, +0.000e+00, +3.688e-02, -8.638e-04, -8.514e-05,
             -2.828e-05, +5.403e-07, +4.390e-07, +1.350e-08, +1.800e-09,
             +0.000e+00, -2.736e-02, -2.977e-04, +8.113e-05, +2.329e-07,
             +8.451e-07, +4.490e-08, -8.100e-09, -1.500e-09, +2.000e-10]

    ah_amp=[ -2.738e-01, -2.837e+00, +1.298e-02, -3.588e-01, +2.413e-02,
            +3.427e-02, -7.624e-01, +7.272e-02, +2.160e-02, -3.385e-03,
            +4.424e-01, +3.722e-02, +2.195e-02, -1.503e-03, +2.426e-04,
            +3.013e-01, +5.762e-02, +1.019e-02, -4.476e-04, +6.790e-05,
            +3.227e-05, +3.123e-01, -3.535e-02, +4.840e-03, +3.025e-06,
            -4.363e-05, +2.854e-07, -1.286e-06, -6.725e-01, -3.730e-02,
            +8.964e-04, +1.399e-04, -3.990e-06, +7.431e-06, -2.796e-07,
            -1.601e-07, +4.068e-02, -1.352e-02, +7.282e-04, +9.594e-05,
            +2.070e-06, -9.620e-08, -2.742e-07, -6.370e-08, -6.300e-09,
            +8.625e-02, -5.971e-03, +4.705e-04, +2.335e-05, +4.226e-06,
            +2.475e-07, -8.850e-08, -3.600e-08, -2.900e-09, +0.000e+00]

    bh_amp=[+0.000e+00, +0.000e+00, -1.136e-01, +0.000e+00, -1.868e-01,
            -1.399e-02, +0.000e+00, -1.043e-01, +1.175e-02, -2.240e-03,
            +0.000e+00, -3.222e-02, +1.333e-02, -2.647e-03, -2.316e-05,
            +0.000e+00, +5.339e-02, +1.107e-02, -3.116e-03, -1.079e-04,
            -1.299e-05, +0.000e+00, +4.861e-03, +8.891e-03, -6.448e-04,
            -1.279e-05, +6.358e-06, -1.417e-07, +0.000e+00, +3.041e-02,
            +1.150e-03, -8.743e-04, -2.781e-05, +6.367e-07, -1.140e-08,
            -4.200e-08, +0.000e+00, -2.982e-02, -3.000e-03, +1.394e-05,
            -3.290e-05, -1.705e-07, +7.440e-08, +2.720e-08, -6.600e-09,
            +0.000e+00, +1.236e-02, -9.981e-04, -3.792e-05, -1.355e-05,
            +1.162e-06, -1.789e-07, +1.470e-08, -2.400e-09, -4.000e-10]

    aw_mean=[+5.640e+01, +1.555e+00, -1.011e+00, -3.975e+00, +3.171e-02,
             +1.065e-01, +6.175e-01, +1.376e-01, +4.229e-02, +3.028e-03,
             +1.688e+00, -1.692e-01, +5.478e-02, +2.473e-02, +6.059e-04,
             +2.278e+00, +6.614e-03, -3.505e-04, -6.697e-03, +8.402e-04,
             +7.033e-04, -3.236e+00, +2.184e-01, -4.611e-02, -1.613e-02,
             -1.604e-03, +5.420e-05, +7.922e-05, -2.711e-01, -4.406e-01,
             -3.376e-02, -2.801e-03, -4.090e-04, -2.056e-05, +6.894e-06,
             +2.317e-06, +1.941e+00, -2.562e-01, +1.598e-02, +5.449e-03,
             +3.544e-04, +1.148e-05, +7.503e-06, -5.667e-07, -3.660e-08,
             +8.683e-01, -5.931e-02, -1.864e-03, -1.277e-04, +2.029e-04,
             +1.269e-05, +1.629e-06, +9.660e-08, -1.015e-07, -5.000e-10]

    bw_mean=[+0.000e+00, +0.000e+00, +2.592e-01, +0.000e+00, +2.974e-02,
             -5.471e-01, +0.000e+00, -5.926e-01, -1.030e-01, -1.567e-02,
             +0.000e+00, +1.710e-01, +9.025e-02, +2.689e-02, +2.243e-03,
             +0.000e+00, +3.439e-01, +2.402e-02, +5.410e-03, +1.601e-03,
             +9.669e-05, +0.000e+00, +9.502e-02, -3.063e-02, -1.055e-03,
             -1.067e-04, -1.130e-04, +2.124e-05, +0.000e+00, -3.129e-01,
             +8.463e-03, +2.253e-04, +7.413e-05, -9.376e-05, -1.606e-06,
             +2.060e-06, +0.000e+00, +2.739e-01, +1.167e-03, -2.246e-05,
             -1.287e-04, -2.438e-05, -7.561e-07, +1.158e-06, +4.950e-08,
             +0.000e+00, -1.344e-01, +5.342e-03, +3.775e-04, -6.756e-05,
             -1.686e-06, -1.184e-06, +2.768e-07, +2.730e-08, +5.700e-09]

    aw_amp=[+1.023e-01, -2.695e+00, +3.417e-01, -1.405e-01, +3.175e-01,
            +2.116e-01, +3.536e+00, -1.505e-01, -1.660e-02, +2.967e-02,
            +3.819e-01, -1.695e-01, -7.444e-02, +7.409e-03, -6.262e-03,
            -1.836e+00, -1.759e-02, -6.256e-02, -2.371e-03, +7.947e-04,
            +1.501e-04, -8.603e-01, -1.360e-01, -3.629e-02, -3.706e-03,
            -2.976e-04, +1.857e-05, +3.021e-05, +2.248e+00, -1.178e-01,
            +1.255e-02, +1.134e-03, -2.161e-04, -5.817e-06, +8.836e-07,
            -1.769e-07, +7.313e-01, -1.188e-01, +1.145e-02, +1.011e-03,
            +1.083e-04, +2.570e-06, -2.140e-06, -5.710e-08, +2.000e-08,
            -1.632e+00, -6.948e-03, -3.893e-03, +8.592e-04, +7.577e-05,
            +4.539e-06, -3.852e-07, -2.213e-07, -1.370e-08, +5.800e-09]

    bw_amp=[+0.000e+00, +0.000e+00, -8.865e-02, +0.000e+00, -4.309e-01,
            +6.340e-02, +0.000e+00, +1.162e-01, +6.176e-02, -4.234e-03,
            +0.000e+00, +2.530e-01, +4.017e-02, -6.204e-03, +4.977e-03,
            +0.000e+00, -1.737e-01, -5.638e-03, +1.488e-04, +4.857e-04,
            -1.809e-04, +0.000e+00, -1.514e-01, -1.685e-02, +5.333e-03,
            -7.611e-05, +2.394e-05, +8.195e-06, +0.000e+00, +9.326e-02,
            -1.275e-02, -3.071e-04, +5.374e-05, -3.391e-05, -7.436e-06,
            +6.747e-07, +0.000e+00, -8.637e-02, -3.807e-03, -6.833e-04,
            -3.861e-05, -2.268e-05, +1.454e-06, +3.860e-07, -1.068e-07,
            +0.000e+00, -2.658e-02, -1.947e-03, +7.131e-04, -3.506e-05,
            +1.885e-07, +5.792e-07, +3.990e-08, +2.000e-08, -5.700e-09]


    pi = 3.141592653590000

#   reference day is 28 January
#   this is taken from Niell (1996) to be consistent
    doy = dmjd  - 44239.0 + 1 - 28

#   degree n and order m
    nmax = 9
    mmax = 9

#   unit vector
    x = math.cos(dlat)*math.cos(dlon)
    y = math.cos(dlat)*math.sin(dlon)
    z = math.sin(dlat)
  
# Legendre polynomials (Cunningham)

    V=[[0. for j in range(nmax+1) ] for i in range(nmax+1)]
    W = [[0. for j in range(mmax + 1)] for i in range(mmax + 1)]

    V[0][0]=1.0
    W[0][0]=0.0
    V[1][0]=z*V[0][0]
    W[1][0]=0.0
    # print(z)


    for n in range(2,nmax+1):
        V[n][0]=((2*n-1) * z * V[n-1][0] - (n-1) * V[n-2][0])  / n
        W[n][0]=0.0


    for m in range(1,nmax+1):
        V[m][m] = (2 * m - 1) * (x * V[m - 1][m - 1] - y * W[m - 1][m - 1])
        W[m][m] = (2 * m - 1) * (x * W[m - 1][m - 1] + y * V[m - 1][m - 1])

        if m<nmax:
            V[m + 1][m] = (2 * m + 1) * z * V[m][m]
            W[m + 1][m] = (2 * m + 1) * z * W[m][m]
        for n in range(m+2,nmax+1):
            V[n][m] = ((2 * n - 1) * z * V[n-1][m] - (n + m - 1) * V[n-2][m]) / (n - m)
            W[n][m] = ((2 * n - 1) * z * W[n-1][m] - (n + m - 1) * W[n-2][m]) / (n - m)
    # print(V)
    # print(W)


    #hydrostatic
    bh = 0.0029
    c0h = 0.062
    if dlat <0.: # southern hemisphere
        phh = pi
        c11h = 0.007
        c10h = 0.002
    else: # northern hemisphere
        phh = 0
        c11h = 0.005
        c10h = 0.001
    ch = c0h + ((math.cos(doy / 365.25 * 2. * pi + phh) + 1) * c11h / 2. + c10h) *(1. - math.cos(dlat))

    ahm = 0.
    aha = 0.
    i = 0
    for n in range(0,nmax+1):
        for m in range(0,n+1):
            ahm = ahm + (ah_mean[i] * V[n][m] + bh_mean[i] * W[n][m])
            # print('ahm',i,n,m,ahm,ah_mean[i] , V[n][m] , bh_mean[i],W[n][m])
            aha = aha + (ah_amp[i] * V[n][m]+ bh_amp[i] * W[n][m])
            # print('aha', i, n, m, aha, ah_amp[i], V[n][m], bh_amp[i], W[n][m])
            i=i+1
    ah=(ahm + aha*math.cos(doy/365.250*2.0*pi))*1.e-5
    # print('ah',ah,ahm,aha,i)
    # print(V)
    # print(W)
    # print(ah_mean)


    # changing this
    #sine = math.sin(pi / 2 - zd)
    #cose = math.cos(pi / 2 - zd)
    #
    sine = np.sin(pi / 2 - zd)
    cose = np.cos(pi / 2 - zd)
    beta = bh / (sine + ch)
    gamma = ah / (sine + beta)
    topcon = (1.0 + ah / (1.0 + bh / (1.0 + ch)))
    gmfh=[0.,0.]
    gmfh[0] = topcon / (sine + gamma)
#   derivative
    gmfh[1] = gmfh[0] ** 2 / topcon * cose * (1 - gamma ** 2 / ah * (1 - beta ** 2 / bh))

#    height correction for hydrostatic mapping function from Niell (1996)
    a_ht = 2.53e-5
    b_ht = 5.49e-3
    c_ht = 1.14e-3
    hs_km = dhgt / 1000.0

    beta = b_ht / (sine + c_ht)
    gamma = a_ht / (sine + beta)
    topcon = (1.0 + a_ht / (1.0 + b_ht / (1.0 + c_ht)))
    ht_corr_coef = 1 / sine - topcon / (sine + gamma)
    ht_corr = ht_corr_coef * hs_km
    gmfh[0] = gmfh[0] + ht_corr

#   derivative
    gmfh[1]= gmfh[1] + (cose/sine**2 - topcon*cose/(sine+gamma)**2*(1-gamma**2/a_ht*(1-beta**2/b_ht))) * hs_km


#   wet
    bw = 0.00146
    cw = 0.04391
    awm = 0.0
    awa = 0.0
    i=0
    for n in range(0,nmax+1):
        for m in range(0,n+1):
            awm = awm + (aw_mean[i] * V[n][m] + bw_mean[i] * W[n][m])
            awa = awa + (aw_amp[i] * V[n][m] + bw_amp[i] * W[n][m])
            i=i+1

    aw = (awm + awa * math.cos(doy / 365.250 * 2 * pi))*1e-5
    beta = bw / (sine + cw)
    gamma = aw / (sine + beta)
    topcon = (1.0 + aw / (1.0 + bw / (1.0 + cw)))

    gmfw = [0., 0.]
    gmfw[0] = topcon / (sine + gamma)

#   derivative
    gmfw[1] = gmfw[0] ** 2 / topcon * cose * (1 - gamma ** 2 / aw * (1 - beta ** 2 / bw))

    return [gmfh[0],gmfh[1],gmfw[0],gmfw[1]]


def sita_Earth(Hr, e_A):                                   
    """
    This function computes the angular separation of the antenna and the 
    reflection point in earth surface, view from earth center

    See Equation (7) in Peng (2023), DOI: 10.1109/TGRS.2023.3332422

    Parameters
    ----------
    Hr : float
        reflector height in meters (height difference between the antenna and the reflecting surface)
    e_A : float
        apparent elevation angle at the antenna, in degrees

    Returns
    -------
    sita_E : float
        earth center angle in degrees

    """

    # earth center angle between the reflection point and the GNSS antenna
    sita_E = Hr / (6378137. * np.tan(np.radians(e_A)))
    # 
    #return round(np.degrees(sita_E), 6)                              # in degree
    return np.degrees(sita_E)                              # in degree

def sita_Satellite(Hr, e_A):                               
    """
    This function computes the angle formed by the antenna-satellite line-of-sight 
    and the reflection point-satellite LoS

    Equation (8) in Peng (2023), DOI: 10.1109/TGRS.2023.3332422

    Parameters
    ----------
    Hr : float
        reflector height in meters (height difference between the antenna and the reflecting surface)
    e_A : float
       apparent elevation angle at the antenna, in degree

    Returns
    -------
    sita_S : float
        satellite angle in degrees (small for MEO satellites)

    """

    ant2satell = 4.*6378137.    # assume satellite distance 4 times earth radius
    sita_S = 2 * Hr * np.cos(np.radians(e_A)) / ant2satell
    #return round(np.degrees(sita_S), 6)    # in degree, small for MEO satellites
    return np.degrees(sita_S)    # in degree, small for MEO satellites

def dH_curve(Hr, Re, e_A):                       
    """
    Computes vertical displacement of the reflection point vs. that of a "planar reflection"

    See Equation (7) in Peng (2023), DOI: 10.1109/TGRS.2023.3332422

    Parameters
    ----------
    Hr : float
        reflector height in meters (height difference between the antenna and the reflecting surface)
    Re : float
        (Gaussian) radius of the Earth in meters
    e_A : float
       apparent elevation angle at the antenna, in degrees

    Returns
    -------
    dH : float
        vertial displacement of the reflection point in meters

    """
    dH = Re * (1. - np.cos(Hr / np.tan(np.radians(e_A)) / Re))
    #return round(dH, 6)    # in meters

    return dH    # in meters

def Hv_Hr_ratio(Hr, Re, e_A):
    """
    This function computes the ratio between the "vertical height difference between
    the antenna and the refection point" and the "reflector height", assuming a 
    sperical reflector (ocean)

    See equation (23) in Peng (2023), DOI: 10.1109/TGRS.2023.3332422

    Parameters
    ----------
    Hr : float
        approximate reflector height in meters 
        (height difference between the antenna and the reflecting surface)
    Re : float
        (Gaussian) radius of the Earth in meters
    e_A : float
       apparent elevation angle at the antenna, in degree

    Returns
    -------
    the_ratio : float
        ratio, allways bigger than 1


    """
    dH = Re * (1. - np.cos(Hr / np.tan(np.radians(e_A)) / Re))
    the_ratio = (Hr+dH)/Hr #ratio, allways bigger than 1

    return the_ratio    

def N_layer(N_antenna, Hr):
    """
    Computes average refractivity of the top (antenna) and bottom (reflecting surface) of this layer

    See Equation (14) in Peng (2023), DOI: 10.1109/TGRS.2023.3332422

    Parameters
    ----------
    N_antenna : float
        refractivity at the antenna in ppm
    Hr : float
        reflector height in meters (height difference between the antenna and the reflecting surface)

    Returns
    -------
    Nl : float
        average refractivity in ppm in this layer

    """
    Nl = N_antenna *(1+np.exp(Hr/8000.)) /2
    #return round(Nl, 4)     #in ppm
    return Nl      #in ppm


def saastam2(press, lat, height):
    """
    This function computes the Zenith Hydrostatic Delay using the Saastamoinen model 
    with updated refractivity equation from Rueger (2002)

    Saastamoinen, J. (1972). Atmospheric corrections for the troposphere and 
    stratosphere in radio ranging of satellites. The Use of Artificial Satellites for 
    Geodesy, Geophysics Monograph Service, 15, 274-251.

    Feng, P., Li, F., Yan, J., Zhang, F., & Barriot, J. P. (2020). 
    Assessment of the accuracy of the Saastamoinen model and VMF1/VMF3 mapping functions 
    with respect to ray-tracing from radiosonde data in the framework of GNSS meteorology. 
    Remote Sensing, 12(20), 3337.

    Parameters
    ----------
    press : float
        atmospheric total pressure in hPa
    lat : float
        latitude of the station, degrees
    height : float
        ellipsoidal height of the station in meters 

    Returns
    -------
    zhd : float
        zenith hystostatic delay in meters

    """
    phi = lat / 180 * 3.14159265359
    height = height / 1000.
    f = 1. - 0.0026 * np.cos(2. * phi) - 0.00028 * height
    # ZHD in meters
    zhd = 2.2794 * press / f / 1000.                

    return zhd

def mpf_tot(gmf_h, gmf_w, zhd, zwd):  
    """
    Finds the total mapping function by weighting the hydrostatic and wet mapping 
    function with the zenith hydrostatic and wet delay. 

    Author: Peng Feng
    
    Parameters
    ----------
    gmf_h : float
        hydrostatic mapping function
    gmf_w : float
        wet mapping function
    zhd : float
        zenith hydrostatic delay in meters
    zwd : float
        zenith wet delay in meters

    Returns
    -------
    mpf_tot1 : float
        total mapping function
    """

    mpf_tot1=(gmf_h*zhd+gmf_w*zwd)/(zhd+zwd)

    return mpf_tot1

def dmpf_dh(ele, dhgt):    
    """
    Station height correction of the hydrostatic mapping function (Niell, 1996)
    This is translated from Johannes Boehm's vmf1_ht.f Fortran code

    Niell, A. E. (1996). Global mapping functions for the atmosphere delay at
    radio wavelengths. Journal of geophysical research: solid earth, 101(B2), 3227-3246.

    Boehm, J., Werl, B., & Schuh, H. (2006). Troposphere mapping functions for
    GPS and very long baseline interferometry from European Centre for Medium‐Range
    Weather Forecasts operational analysis data. JGR: Solid Earth, 111(B2).

    Parameters
    ----------
    ele : float
        true elevation angle in degree
    dhgt : float
        height difference in meters. In GNSS-IR, this is reflector height; 
        in applying mapping function grid products, this is the height 
        difference between the antenna and the grid point height

    Returns
    -------
    ht_corr : float
        correction to the hydrostatic mapping function (vmf1h= vmf1h + ht_corr)

    """
    sine = np.sin(np.radians(ele))
    [a_ht, b_ht, c_ht] = [0.0000253, 0.00549, 0.00114]
    hs_km = dhgt / 1000.
    beta = b_ht / (sine + c_ht)
    topcon = (1. + a_ht / (1. + b_ht / (1. + c_ht)))
    ht_corr_coef = 1 / sine - topcon / (sine + a_ht / (sine + beta))

    return ht_corr_coef * hs_km


def Ulich_Bending_Angle_original(ele, N0):   
    """
    This function computes the atmospheric bending angle with the Ulich equation.

    Equation (18) in Ulich, B. L. (1981). Millimeter wave radio telescopes: 
    Gain and pointing characteristics. International Journal of Infrared and Millimeter Waves, 2, 293-310.

    Parameters
    ----------
    ele : float
        true elevation angle in degrees
    N0 : float
        refractivity in part-per-million

    Returns
    ----------
    dele : float
        bending angle (angular difference between apparent and true elevation angle), in degrees

    """
    ele = np.radians(ele)
    r = N0/1000000.
    f = np.cos(ele) / (np.sin(ele) + 0.00175 * np.tan(np.radians(87.5) - ele))

    return np.degrees(r * f)


def Equivalent_Angle_Corr_mpf(ele, mpf_tot, N0, Hr_apr):
    """
    This function computes the "equvilent" angular correction to apply the 
    tropospheric delay calculated with the mapping function.


    See: Williams, S. D. P., & Nievinski, F. G. (2017). Tropospheric delays in ground‐based 
    GNSS multipath reflectometry—Experimental evidence from coastal sites. Journal of Geophysical 
    Research: Solid Earth, 122(3), 2310-2327.

    Strandberg, J. (2020). New methods and applications for interferometric GNSS 
    reflectometry. Chalmers Tekniska Hogskola (Sweden).

    Parameters
    ----------
    ele : float
        true elevation angle in degrees
    mpf_tot : float
        total mapping function, units?
    N0 : float
        refractivity at GNSS antenna in part-per-million
    Hr_apr : float
        approximate reflector height in meters

    Returns
    -------
    dele : float
        equvilent angular correction in degrees
    """

    Nl = N_layer(N0, Hr_apr)
    dsin_mpf = Nl * mpf_tot / 1000000.  # correction to the sine value
    ele_eqv = np.degrees(np.arcsin(np.sin(np.radians(ele)) + dsin_mpf))
    dele = ele_eqv -ele

    return dele

def asknewet(e, Tm, lambda_val):
    """
    Determines the zenith wet delay based on the equation 22 by Askne and Nordius (1987)

    Askne and Nordius, Estimation of tropospheric delay for microwaves from surface weather data,
    Radio Science, Vol 22(3): 379-386, 1987.

    Source: Peng Feng
   
    Parameters
    ----------
    e : float
        water vapor pressure in hPa
    Tm : float
        mean temperature in Kelvin
    lambda_val: float
        water vapor lapse rate (see definition in Askne and Nordius 1987)

    Returns
    -------
    zwd :  float
        zenith wet delay in meter

    """

    #
    # Example 1:
    # e =  10.9621 hPa
    # Tm = 273.8720
    # lambda_val = 2.8071
    #
    # output:
    # zwd = 0.1176 m
    # translated from matlab code available at https://vmf.geo.tuwien.ac.at/codes/
    # Johannes Boehm, 3 August 2013
    # ---

    # coefficients
    k1 = 77.604  # K/hPa
    k2 = 64.79  # K/hPa
    k2p = k2 - k1 * 18.0152 / 28.9644  # K/hPa
    k3 = 377600  # KK/hPa

    # mean gravity in m/s**2
    gm = 9.80665
    # molar mass of dry air in kg/mol
    dMtr = 28.965 * 1e-3
    # universal gas constant in J/K/mol
    R = 8.3143

    # specific gas constant for dry consituents
    Rd = R / dMtr

    zwd = 1e-6 * (k2p + k3 / Tm) * Rd / (lambda_val + 1) / gm * e

    return zwd




import argparse
import json
import os
import sys


import gnssrefl.spline_functions as spline_functions
import gnssrefl.refraction as refr

from gnssrefl.utils import str2bool


def parse_arguments():
    parser = argparse.ArgumentParser()
    parser.add_argument("station", help="station name", default=None, type=str)
    parser.add_argument("year", default=None, type=int, help="year")
    parser.add_argument("doy", default=None, type=int, help="doy")
    parser.add_argument("signal", default=None, type=str, help="L1, L2, L5, L1+L2, L1+L2+L5, etc")

    parser.add_argument("-peak2noise", default=None, type=float, help="Peak2noise ratio for Quality Control, default now 2.5")
    parser.add_argument("-constel", default=None, type=str, help="Only a single constellation (G,E, or R)")
    parser.add_argument("-screenstats", default=None, type=str, help="screen stats, False is default")
    parser.add_argument("-dec", default=None, type=int, help="SNR file decimator (seconds)")
    parser.add_argument("-polydeg", default=None, type=int, help="polynomial degree for direct signal removal (default is 2)")
    parser.add_argument("-snrfit", default=None, type=str, help="Do invsnr fit? True is the default ")
    parser.add_argument("-plt", default=None, type=str, help="Plot to the screen?  default is True")
    parser.add_argument("-doy_end", default=None, type=str, help="day of year to end analysis")
    parser.add_argument("-lspfigs", default=None, type=str, help="Make LSP plots, default False.")
    parser.add_argument("-snrfigs", default=None, type=str, help="Make SNR plots, default False.")
    parser.add_argument("-knot_space", default=None, type=int, help="knot spacing in hours (default is 3)")
    parser.add_argument("-rough_in", default=None, type=str, help="Roughness (default is 0.1)")
    parser.add_argument("-risky", default=None, type=str, help="Risky taker related to gaps/knot spacing, False is default.")
    parser.add_argument("-snr", default=None, type=int, help="SNR file ending. Default is 66")
    parser.add_argument("-outfile_type", default=None, type=str, help="Output file type (txt or csv)")
    parser.add_argument("-outfile_name", default=None, type=str, help="Output file name")
    parser.add_argument("-outlier_limit", default=None, type=str, help="outliers limit (m)")
    parser.add_argument("-no_dots", default=None, type=str, help="bool, no lomb scargle  results plotted")
    parser.add_argument("-delta_out", default=None, type=str, help="Output increment, in seconds (default is 300)")
    parser.add_argument("-refraction", default=None, type=str, help="bool, Set to False to turn off")
    parser.add_argument("-json_override", default=None, type=str, help="bool, Override json file name")
    parser.add_argument("-lastday_seconds", default=None, type=int, help="if last doy incomplete, provides last timepoint, this option currently does not work")

    args = parser.parse_args().__dict__

    # convert all expected boolean inputs from strings to booleans
    boolean_args = ['screenstats', 'snrfit', 'plt', 'lspfigs', 'snrfigs', 'risky', 'no_dots','refraction', 'json_override',]
    args = str2bool(args, boolean_args)

    # only return a dictionary of arguments that were added from the user - all other defaults will be set in code below
    return {key: value for key, value in args.items() if value is not None}


def invsnr(station: str, year: int, doy: int, signal: str, peak2noise: float = 2.5, constel: str = None, 
        screenstats: bool = False, dec: int = 1, polydeg: int = 2, snrfit: bool = True, plt: bool = True,
        doy_end: int = None, lspfigs: bool = False, snrfigs: bool = False, knot_space: int = 3, 
        rough_in: float = 0.1, risky: bool = False, snr: int = 66, outfile_type: str = 'txt', 
        outfile_name: str = '', outlier_limit: float = 0.5, no_dots: bool = False, delta_out: int = 300, 
           refraction: bool = True, json_override: bool = False, lastday_seconds: int=0):
    """
    You must have run invsnr_input before using this code. This is the wrapper code that does the 
    invsnr modelling. Note: outfile_name and outfile_type are unnecessary. Consolidate them.

    In an earlier version of the code the pk2nlim was set to 4.  Later the definition of the metric was changed,
    and this made the default setting far too stringent. In short, no arcs were being found.  As of 2023/10/28 it
    is set to 2.5.  This may not be optimal, but it is not as bad as 4.  Please set it yourself as you prefer.
    Note: it will not the same as gnssir as this code was written separately and for a different purpose.

    pktnlim is now known externally as peak2noise
    
    Examples
    --------
    invsnr sc02 2023 15 L1+L2+L5 
        would analyze day of year 15 and the L1, L2, and L5 signals
    invsnr sc02 2023 15 ALL
        would analyze day of year 15 and all signals
    invsnr sc02 2023 15 L1+L2
        would analyze day of year 15 and just L1 and L2
    invsnr sc02 2023 15 L1+L2 -pk2nlim 2
        would analyze day of year 15 and just L1 and L2, lower peak to noise limit ratio
    invsnr sc02 2023 15 L1+L2  -doy_end 18
        would analyze day of years 15 through 18 and L1 and L2 signals

    Parameters
    ----------
    station : str
        four character ID 
    year : int
        Year
    doy : int
        Day of year
    signal : str
        signal to use, L1  L2 L5 L6 L7 L1+L2 L1+L2+L5 L1+L5 ALL
    peak2noise: float, optional
        Peak2noise ratio limit for Quality Control.
        Default is 2.5
    constel: str, optional
        Only a single constellation.
        Default is gps, glonass, and galileo.
        value options:

            G : GPS

            E : Galileo

            R : Glonass

            C : Beidou

            withBeidou : adds Beidou to the default.

    screenstats: bool, optional
        Whether to print out stats to the screen.
        Default is False
    dec : int, optional
        SNR file decimator (seconds)
        Default is 1 (everything)
    polydeg : integer, optional
        polynomial degree for direct signal removal
        Default is 2
    snrfit : bool, optional
        Whether to do the inversion or not
        Default is True
    plt : bool, optional
        Whether to plot to the screen or not
        Default is True
    doy_end : int, optional
        day of year to end analysis.
        Default is None.
    lspfigs : bool, optional
        Whether or not to make LSP plots
        Note: Don't turn these on unless you really need plots because it is 
        slow to make one per satellite arc.
        Default is False
    snrfigs : boolean, optional
        Whether or not to make SNR plots
        Don't turn these on unless you really need plots because it is 
        slow to make one per satellite arc.
        Default is False
    knot_space : float, optional
        Knot spacing in hours
        Default is 3
    rough_in : float, optional
        Roughness
        Default is 0.1
    risky : bool, optional
        Risky taker related to gaps/knot spacing
        Default is False
    snr : int, optional
        SNR file ending. Default is 66
    outfile_type : string, optional
        output file type, txt or csv
        Default is txt
    outfile_name : string, optional
        output file name.
        Default is ??
    outlier_limit : float, optional
        Outliers plotted. (meters)
        Default is 0.5
    no_dots : bool, optional
        To plot lombscargle or not.
        Default is False
    delta_out : int, optional
        Output increment, in seconds.
        Default is 300
    refraction : bool, optional
        Default is True
    json_override : bool, optional
        Override json file name
        Default is False
    lastday_seconds : int, optional
        last time point (seconds of the day)
        should really be read from the file - 
        if you don't provide this the fit blows up

    """

    if len(station) != 4:
        print('Stations must be four characters long. Exiting.')
        sys.exit()

    if signal.upper() not in ['L1', 'L2', 'L5', 'L6', 'L7', 'L1+L2', 'L1+L2+L5', 'L1+L5', 'ALL']:

        print('Currently only allow L1,L2,L5,L6,L7 and various combinations')
        print('Illegal signal:', signal)
        sys.exit()

    if signal.upper() == 'ALL':
        print('Using all signals')
        signal = 'L1+L2+L5+L6+L7'
    print(signal)

    # build a dictionary with the analysis inputs
    # environment variable for the file inputs
    xdir = os.environ['REFL_CODE']

    # location of the analysis inputs, if it exists
    jsondir  = xdir + '/input/'
    instructions2 =  jsondir + station + '.inv.json'

    if json_override is False:
        if os.path.isfile(instructions2):
            #print('using:', instructions2)
            with open(instructions2) as f:
                lsp = json.load(f)
        else:
            print('Instruction file does not exist.', instructions2, ' Exiting.')
            sys.exit()
    else:
        instructions2 = json_override
        if os.path.isfile(instructions2):
            #print('using:', instructions2)
            with open(instructions2) as f:
                lsp = json.load(f)
        else:
            print('Instruction file does not exist.', instructions2, ' Exiting.')
            sys.exit()


# save json inputs to variables
# this is ridiculous - I should just send the dictionary!
    precision = lsp['precision']
    azilims = lsp['azilims']
    elvlims = lsp['elvlims']
    rhlims = lsp['rhlims']
    l2c_only = lsp['l2c_only']
    if 'pktnlim' in lsp:
        print('use peak2noise from json')
        peak2noise = lsp['pktnlim']


# multi doy listing
    if doy_end is None:
        doy_end = doy
    else:
        doy_end = int(doy_end)
    
    if (doy_end < doy):
        print('doy_end cannot be less than doy')
        sys.exit()

# set constellation.
    if constel is None:
        # added beidou 22feb09
        #satconsts=['E','G','R','C'] # the default is gps,glonass, and galileo
        satconsts=['E', 'G', 'R'] # the default is gps, glonass, and galileo
    else: 
        satconsts=[constel]
        # save people from themselves
        if (constel == 'R') and (signal == 'L5'):
            print('illegal constellation/frequency choice', constel, '/', 'signal')
            sys.exit()
        if (constel == 'E') and (signal == 'L2'):
            print('illegal constellation/frequency choice', constel, '/', signal)
            sys.exit()
        if (constel == 'C') and (signal == 'L1' or signal == 'L5'):
            print('illegal constellation/frequency choice', constel, '/', signal)
            sys.exit()

    if (constel == 'withBeidou'):
        satconsts=['E', 'G', 'R', 'C'] # the default is gps, glonass, and galileo
    if (constel == 'GPS+Gal'):
        satconsts=['E', 'G'] # 

    print('Roughness:', rough_in)

    print('Knot spacing (hours):', knot_space)

    kdt = knot_space * 60 * 60  # change knot spacing to seconds

    # trying to add refraction to an existing dictionary. this is how it is done in the main lsp code
    lsp['refraction'] = False

    if refraction:
        lsp['refraction'] = True

    spline_functions.snr2spline(station, year, doy, azilims, elvlims, rhlims, precision, kdt, signal=signal,
                                lspfigs=lspfigs, snrfigs=snrfigs, snrfit=snrfit, doplot=plt, pktnlim=peak2noise,
                                satconsts=satconsts, screenstats=screenstats, tempres=dec, doy_end=doy_end,
                                l2c_only=l2c_only, rough_in=rough_in, risky=risky, snr_ending=snr,
                                outfile_type=outfile_type, delta_out=delta_out, lsp=lsp, outfile_name=outfile_name,
                                outlier_limit=outlier_limit, no_dots=no_dots,lastday_seconds=lastday_seconds)


def main():
    args = parse_arguments()
    invsnr(**args)


if __name__ == "__main__":
    main()





import json
import numpy as np
import os
import platform
import warnings

from enum import Enum
from typing import get_type_hints
from pathlib import Path


def validate_input_datatypes(obj, **kwargs):
    hints = get_type_hints(obj)

    # iterate all type hints
    for attr_name, attr_type in hints.items():
        if attr_name == 'return':
            continue
        try:
            if kwargs[attr_name] is None:
                pass
            elif not isinstance(kwargs[attr_name], attr_type):
                raise TypeError('Argument %r is not of type %s' % (attr_name, attr_type))
        except KeyError:
            pass


def str2bool(args, expected_bools):
    if type(expected_bools) is str:
        expected_bools = [expected_bools]
    elif type(expected_bools) is list:
        pass
    else:
        print('str2bool accepts a string or a list of strings')

    for string in expected_bools:
        if isinstance(args[string], bool):
            pass
        elif args[string] is None:
            pass
        elif args[string].lower() in ('yes', 'true', 't', 'y', '1'):
            args[string] = True
        elif args[string].lower() in ('no', 'false', 'f', 'n', '0'):
            args[string] = False

    return args


class FileTypes(str, Enum):
    """
    Files to either read from or save to.
    """
    apriori_rh_file = "apriori_rh_file"
    daily_avg_phase_results = "daily_avg_phase_results"
    make_json = "make_json"
    phase_file = "phase_file"
    volumetric_water_content = "volumetric_water_content"
    directory = "directory"


# TODO we should do something like this below for all of our file structuring so it's all in one place
# If we decide to change directories paths etc later then we would only need to change it here and nowhere else
class FileManagement:
    """
    FileManagement is designed to easily read the files that this package relies on.
    Required parameters include station and file_type from FileTypes class.
    Optional parameters are year, doy, and file_not_found_ok.
    """

    def __init__(self, station, file_type: FileTypes, year: int = None, doy: int = None, file_not_found_ok: bool = False):
        self.station = station
        self.file_type: FileTypes = file_type
        self.year = year
        self.doy = doy
        self.file_not_found_ok = file_not_found_ok

        self.xdir = Path(os.environ["REFL_CODE"])

    def get_file_path(self):
        """
        Get the path of a specific file from the FileTypes class.
        Returns file paths requested as a string
        """
        if self.file_type in FileTypes.__dict__.keys():
            files = {FileTypes.apriori_rh_file: self.xdir / "input" / f"{self.station}_phaseRH.txt",
                     FileTypes.daily_avg_phase_results: self.xdir / "Files" / f"{self.station}_phase.txt",
                     FileTypes.make_json: self.xdir / "input" / f"{self.station}.json",
                     FileTypes.volumetric_water_content: self.xdir / "Files" / f"{self.station}_vwc.txt"
                     }

            if self.year and self.doy:
                files[FileTypes.phase_file] = self.xdir / str(self.year) / 'phase' / str(self.station) / f'{self.doy:03d}.txt'

            return files[self.file_type]
        else:
            raise ValueError("The file type you requested does not exist")

    def read_file(self, transpose=False, **kwargs):
        """
        Reads the requested file amd returns results of file as an array.
        Can use transpose parameter to transpose the results.
        """
        file_path = self.get_file_path()
        if file_path.exists():
            pass
        else:
            if not self.file_not_found_ok:
                raise FileNotFoundError(f"{file_path} does not exist.")
        if file_path.suffix == ".json":
            with open(file_path, 'r') as my_json:
                return json.load(my_json)
        else:
            if file_path.stat().st_size > 0:
                result = np.genfromtxt(file_path, **kwargs)

                if transpose is False:
                    return result
                else:
                    return result.T


def read_files_in_dir(directory, transpose=False):
    """
    Read all files in a given directory. Directory given must be an absolute path.
    Returns an n-d array of results.
    Can use optional parameter transpose to transpose the results.
    """
    dir_path = Path(directory)
    if dir_path.is_absolute():
        # Make sure directory exists
        if dir_path.exists():
            # Make sure given path is a directory - not a file
            if dir_path.is_dir():
                data = []
                #result_files = list(dir_path.glob("*.txt"))
                result_files = list(dir_path.glob("???.txt"))
                if result_files is not None:
                    for file in result_files:

                        with warnings.catch_warnings():
                            warnings.simplefilter("ignore")
                            file_data = np.genfromtxt(file, comments='%')

                        if (len(file_data) > 0):
                            if len(np.shape(file_data)) == 1:
                                data.append(file_data)
                            else:
                                data.extend(file_data)
                    if data is not None:
                        if transpose:
                            data = np.array(data)
                            return data.T
                        else:
                            return data
                    else:
                        print(f"Files returned no data")
                else:
                    print(f"There are no files to read in this directory: {dir_path}")
            else:
                raise NotADirectoryError(f"path is not a directory. {dir_path}")
        else:
            print(f"Directory does not exist: {dir_path}")
    else:
        raise ValueError(f"path is not absolute. Please provide an absolute path: {dir_path}")

def check_environment():
    try:
        os.environ['ORBITS']
        os.environ['REFL_CODE']
        os.environ['EXE']

        environment_set = True
    except KeyError:
        environment_set = False

    return environment_set


def set_environment(refl_code, orbits, exe):
    os.environ['ORBITS'] = str(Path(orbits))
    os.environ['REFL_CODE'] = str(Path(refl_code))
    os.environ['EXE'] = str(Path(exe))
    print('environment variable ORBITS set to path', os.environ['ORBITS'],
          '\nenvironment variable REFL_CODE set to path', os.environ['REFL_CODE'],
          '\nenvironment variable EXE set to path', os.environ['EXE'])


def get_sys():
    system = platform.platform().lower()
    valid_os = ['linux64', 'macos']

    for os in valid_os:
        if os in system:
            if os == 'macos':
                if platform.processor() == 'arm':
                    os = 'mac-newchip'
            return os



# -*- coding: utf-8 -*-
import argparse
import subprocess
import gnssrefl.gps as g

def main():
    """
    checks the first 100 lines of a rinex file in a quest to determine
    the approximate Cartesian coordinates, which are also displayed 
    on the screen in LLH. It does not care if it is rinex, hatanaka rinex,
    or rinex3. But it does require it to not be gzipped or unix compressed.
    If you would like that option, please submit a PR.

    Example
    -------
    rinex_coords p1013550.22o

    """

    parser = argparse.ArgumentParser()
    parser.add_argument("rinexfile", help="filename ", type=str)

    args = parser.parse_args()


    fname = args.rinexfile
    cmd = 'head -100 ' + fname  + ' | grep POS'
    line=subprocess.getoutput(cmd)
    x= float(line.split()[0])
    y= float(line.split()[1])
    z= float(line.split()[2])
    xyz = [x,y,z]

    print("Cartesian (m) %15.4f %15.4f %15.4f " % (x, y, z) )
    lat,lon,h = g.xyz2llhd(xyz)
    print("Lat Lon Ht (deg deg m) %14.9f %14.9f %9.3f " % (lat, lon, h) )

if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
import argparse
import gnssrefl.gps as g

# this requires python 3.8
#from importlib.metadata import version

def main():
    """
    converts year month day to day of year and prints it to the screen

    MJD is an optional output

    Parameters
    ----------
    year : int
        4 ch year

    month : int
        calendar month

    day : int
        calendar day

    mjd : str
        use T or True to get MJD printed to the screen

    Returns
    -------
    doy : str 
        three character day of the year
    """

    parser = argparse.ArgumentParser()
    parser.add_argument("year", help="year ", type=int)
    parser.add_argument("month", help="month", type=int)
    parser.add_argument("day", help="day", type=int)
    # this is not our normal way to input booleans but I am busy ...!
    parser.add_argument("-mjd", help="Print MJD to the screen. Default is F", type=str,default=None)

    args = parser.parse_args()
    year = args.year
    month = args.month
    day = args.day

    doy,cdoy,cyyyy,cyy = g.ymd2doy(year, month, day )
    print(cdoy)
    if args.mjd is not None:
        if (args.mjd == 'T') or (args.mjd == 'True'):
            mjd = g.getMJD(year,month,day,0)
            print('MJD', mjd)

if __name__ == "__main__":
    main()



import datetime
import matplotlib.pyplot as plt
import numpy as np
import os
from scipy.interpolate import interp1d
import subprocess
import sys
import time


# progress bar for RINEX translation/orbits
from progress.bar import Bar

# my gps libraries
import gnssrefl.gps as g
import gnssrefl.rinpy as rinpy
import gnssrefl.karnak_libraries as k
import gnssrefl.highrate as ch

# fortran codes for translating RINEX
import gnssrefl.gpssnr as gpssnr
import gnssrefl.gnsssnr as gnsssnr
import gnssrefl.gnsssnrbigger as gnsssnrbigger

class constants:
    omegaEarth = 7.2921151467E-5 #      %rad/sec
    mu = 3.986005e14 # Earth GM value
    c= 299792458 # m/sec

#
#
def quickname(station,year,cyy, cdoy, csnr):
    """
    creates filename for a local SNR file

    Parameters
    ----------
    station : str
        station name, 4 character
    year : int
        full year
    cyy : str
        two character year
    cdoy : str
        three character day of year
    csnr : str
        snr ending, i.e. '66' or '99'

    Returns
    -------
    fname : str
        full filename including the directory

    """
    xdir  = os.environ['REFL_CODE'] + '/'
    fname =  xdir + str(year) + '/snr/' + station + '/' + station + cdoy + '0.' + cyy + '.snr' + csnr

    return fname

def run_rinex2snr(station, year, doy,  isnr, orbtype, rate,dec_rate,archive, nol,overwrite,translator,srate, 
                  mk, stream,strip,bkg,screenstats,gzip,timeout):
    """
    main code to convert RINEX files into SNR files 
    now works on a single year and doy

    Parameters
    ----------
    station : str
        4 or 9 character station name. 6 ch allowed for japanese archive
        9 means it is a RINEX 3 file

    year : int
        full year 
    doy : int 
        day of year
    isnr : int
        SNR file type choice, i.e. 66, 88
    orbtype : str
        orbit type, e.g. nav, rapid, gnss
    rate : str
        general sample rate.
        high: use 1-Hz area in the archive
        low: use default area in the archive
    dec_rate : integer
         decimation value
    archive : str
        choice of GNSS archive, of which there are many
    nol: bool
        True: assumes RINEX files are in local directory
        False (default): will look at multiple - or specific archive
    overwrite: bool
        False (default): if SNR file exists, SNR file not made
        True: make a new SNR file
    translator : str
        hybrid (default), fortran, or python
        hybrid uses fortran within he python code
    srate : int
        sample rate for RINEX 3 files
    mk : bool
        makan option
    strip : bool
        reduces observables to only SNR (too many observables, particularly in RINEX 2 files
        will break the RINEX translator)
    bkg : str
        location of bkg files, EUREF or IGS
    screenstats: bool
        whether print statements come to screen
    gzip: bool
        whether SNR files are gzipped after creation
    timeout : int
        optional parameter I am testing out for requests timeout parameter
        in seconds
    """
    #

    NS = len(station)
    if (NS == 4):
        version = 2
        if not mk:
            station = station.lower()
    elif (NS == 6 and archive == 'jp') :
        version = 2
        if not mk:
            station = station[-4:].upper()
    elif (NS == 9):
        #print('Assume RINEX 3');
        version = 3
        station9ch = station.upper()
        if not mk:
            station = station[0:4].lower()
        else:
            station = station[0:4].upper()
    else:
        print('Illegal station input - Station must have 4,6,or 9 characters. Exiting')
        sys.exit()


# old loop thru years and days - will fix later. now avoiding retyping
    if True:
        ann = g.make_nav_dirs(year)
        cyyyy = str(year)
        dec31 = g.dec31(year)
        if True:
            csnr = str(isnr)
            cdoy = '{:03d}'.format(doy)
            if (year<2000):
                cyy = '{:02d}'.format(year-1900)
            else:
                cyy = '{:02d}'.format(year-2000)
            # first, check to see if the SNR file exists
            fname =  quickname(station,year,cyy,cdoy,csnr)
            if screenstats:
                print(fname)
            # now it unzips if that version exists
            snre = g.snr_exist(station,year,doy,csnr)
            if snre:
                if overwrite:
                    print('SNR file exists/you requested overwriting, existing file will be deleted')
                    if os.path.isfile(fname):
                        subprocess.call(['rm', fname]); 
                    if os.path.isfile(fname + '.gz'):
                        subprocess.call(['rm', fname + '.gz']); 
                    snre = False
                else:
                    print('SNR file already exists', fname, '\n')
            else:
                print('SNR file does not already exist. Which means I will try to make it.')


            illegal_day = False
            if (doy > dec31):
                illegal_day = True

            if (not illegal_day) and (not snre):
                r = station + cdoy + '0.' + cyy + 'o'
                rgz = station + cdoy + '0.' + cyy + 'o.gz'
                localpath2 =  os.environ['REFL_CODE'] + '/' + cyyyy + '/rinex/' + station + '/'
                if screenstats:
                    print(localpath2)
                if nol:
                    current_local = os.getcwd()
                    print('Will first assume RINEX file ', station, ' year:', year, ' doy:', doy, 'is located here :', current_local)
                    # this assumes RINEX file is in local directory or "nearby"
                    if version == 2:

                        if mk:
                            the_makan_option(station,cyyyy,cyy,cdoy) # looks everywhere in your local directories
                        if not os.path.exists(r):
                            print('Did not find the plain observation file, so now trying other names/directories')
                            # could try this way? - look for file in localpath2. gunzip if necessary
                            allgood = get_local_rinexfile(r,localpath2)
                        if os.path.exists(r):
                            if screenstats:
                                print('Found the RINEX 2.11 file', r)
                            try:
                                rinext =float(np.loadtxt(r,usecols=0,dtype='str',max_rows=1))
                                print('Apparent Rinex version', rinext)
                                if (rinext != 2.11):
                                    print('Your file is not RINEX v2.11 which is what you told the code it was.')
                                    #sys.exit()

                            except:
                                print('I had problems confirming RINEX version. Will ignore that for now.')

                            if strip:
                                if screenstats:
                                    print('Testing out stripping the RINEX 2 file here')
                                k.strip_rinexfile(r)
                            conv2snr(year, doy, station, isnr, orbtype,rate,dec_rate,archive,translator)

                        else:
                            print('You Chose the No Look Option, but did not provide the needed RINEX file.')
                    if version == 3:
                        if rate == 'high':
                            csrate = '01' # high rate assumes 1-sec
                        else:
                            csrate = '{:02d}'.format(srate)
                        streamid = '_' + stream  + '_'
                        # this can be done in a function now ...
                        r3cmpgz = station9ch + streamid + str(year) + cdoy + '0000_01D_' + csrate + 'S_MO.crx.gz'
                        r3 = station9ch + streamid + str(year) + cdoy + '0000_01D_' + csrate + 'S_MO.rnx'
                        r3gz = station9ch + streamid + str(year) + cdoy + '0000_01D_' + csrate + 'S_MO.rnx.gz'
                        r2 = station + cdoy + '0.' + cyy + 'o'
                        if os.path.exists(r3cmpgz):
                            if screenstats: 
                                print('Try to translate', r3cmpgz)
                            deletecrx = True
                            translated, rnx_filename = go_from_crxgz_to_rnx(r3cmpgz,deletecrx)
                        if os.path.exists(r3gz):
                            if screenstats: 
                                print('Try to gunzip ', r3gz)
                            subprocess.call(['gunzip', r3gz])

                        # have not found the rinex 3 file
                        if not os.path.exists(r3):
                            local_rinex3_dir  = os.environ['REFL_CODE'] + '/' + cyyyy + '/rinex/' + station + '/'
                            print('try looking for RINEX 3 in ', local_rinex3_dir)
                            lrinex3 = local_rinex3_dir+r3
                            if os.path.exists(lrinex3):
                                subprocess.call(['cp', lrinex3, '.'])
                            else:
                                lrinex3 = local_rinex3_dir+r3cmpgz
                                if os.path.exists(lrinex3):
                                    subprocess.call(['cp', lrinex3, '.']); 
                                    deletecrx = True
                                    translated, rnx_filename = go_from_crxgz_to_rnx(r3cmpgz,deletecrx)
                        if os.path.exists(r3):
                            rinext =float(np.loadtxt(r3,usecols=0,dtype='str',max_rows=1))
                            print('Apparent Rinex version', rinext)
                            if (rinext < 3):
                                print('Your file is not RINEX v3 or higher which is I was expecting. Exiting.')
                                sys.exit()
                            if screenstats: 
                                print('The RINEX 3 file exists locally', r3)
                            # convert to RINEX 2.11
                            fexists = g.new_rinex3_rinex2(r3,r2,dec_rate)
                            if fexists:
                                conv2snr(year, doy, station, isnr, orbtype,rate,dec_rate,archive,translator)
                            else:
                                print('Something about the RINEX 3-2 conversion did not work')
                        else:
                            print('You Chose the No Look Option, but did not provide the needed RINEX3 file ', r3)
                            print('I looked for files ending with rnx, rnx.gz, and crx.gz in the local directory')
                            print('I looked for files ending with rnx and crx.gz in $REFL_CODE/YYYY/rinex for your station')

                else:
                    if screenstats:
                        print('Will seek the RINEX file from an external archive')
                    if version == 3:
                        fexists = False
                        rnx_filename = '' # just in  case?
                        print(station9ch, ' year:', year, ' doy:', doy, 'from: ', archive)
                        r2 = station + cdoy + '0.' + cyy + 'o'
                        rinex2exists = False; rinex3name = '';
                        if (rate == 'high'):
                            print('This code only accesses 1-Hz Rinex 3 data at BKG, CDDIS, GA, the Spanish IGN, and GNET')
                            if archive == 'ga':
                                deleteOld = True
                                # this downloads RINEX 3 and converts to Rinex 2
                                r2, fexists= g.ga_highrate(station9ch,year,doy,dec_rate,deleteOld)
                                if screenstats:
                                    print('RINEX 2 file derived from the GA archive should now exist:', r2)
                            if archive == 'gnet':
                                rnx_filename,foundit = g.greenland_rinex3(station9ch, year, doy )
                                print(rnx_filename,foundit)
                                if foundit: 
                                    print('The RINEX 3 file has been downloaded and gunzipped. Try to make ', r2)
                                    fexists = g.new_rinex3_rinex2(rnx_filename,r2,dec_rate)

                            if archive == 'cddis':
                                bad_day = g.cddis_restriction(year, doy,'cddis')
                                if not bad_day:
                                    rnx_filename,foundit = ch.cddis_highrate(station9ch, year, doy, 0,stream,dec_rate)
                                else: 
                                    print('Check for the tar version')
                                    rnx_filename,foundit = ch.cddis_highrate_tar(station9ch, year, doy, 0,stream,dec_rate)
                                    print(rnx_filename, ' returned from tar version')
                                if not foundit:
                                    print('No high-rate RINEX data will be downloaded')
                                    foundit = False; fexists = False; rnx_file = ''
                                else:
                                    if screenstats:
                                        print('The RINEX 3 file has been downloaded from CDDIS. Now try to make ', r2)
                                    fexists = g.new_rinex3_rinex2(rnx_filename,r2,dec_rate)
                            if archive == 'bkg':
                                # this is confusing - so the bkg variable is either IGS or EUREF
                                bad_day = g.cddis_restriction(year, doy,'bkg')
                                if not bad_day:
                                    rnx_filename,foundit = ch.bkg_highrate(station9ch, year , doy, 
                                                                           0,stream,dec_rate,bkg,timeout=timeout)
                                else:
                                    print('Will try the tar version ')
                                    rnx_filename,foundit = ch.bkg_highrate_tar(station9ch, year, doy, 
                                                                               0,stream,dec_rate,bkg,timeout=timeout)
                                if foundit:
                                    if screenstats:
                                        print('The RINEX 3 file has been downloaded from the BKG and merged. Now try to make ', r2)
                                    fexists = g.new_rinex3_rinex2(rnx_filename,r2,dec_rate)
                            if archive == 'ignes':
                                bad_day = g.cddis_restriction(year, doy,'bkg')
                                if not bad_day:
                                    rnx_filename,foundit = ch.esp_highrate(station9ch, year, doy, 0,stream,dec_rate)
                                else:
                                    print('No high-rate RINEX data will be downloaded')
                                    foundit = False; fexists = False; rnx_file = ''
                                if foundit:
                                    if screenstats:
                                        print('The RINEX 3 file has been downloaded from IGN ES and merged. Try to make ', r2)
                                    fexists = g.new_rinex3_rinex2(rnx_filename,r2,dec_rate)

                        else:
                            if (archive == 'all'):
                                file_name,foundit = k.universal_all(station9ch, year, doy,srate,stream,screenstats)
                                if (not foundit): # try again
                                    file_name,foundit = k.universal_all(station9ch, year, doy, srate,k.swapRS(stream),screenstats)
                            else:
                                #print('stream',stream)
                                file_name,foundit = k.universal(station9ch, year, doy, archive,srate,stream)
                                if (not foundit): # try again
                                    #print('stream',stream)
                                    file_name,foundit = k.universal(station9ch, year, doy, archive,srate,k.swapRS(stream))
                            if foundit: # version 3 found - now need to gzip, then hatanaka decompress
                                deletecrx = True # no point keeping this around
                                translated, rnx_filename = go_from_crxgz_to_rnx(file_name,deletecrx)
                            # now make rinex2
                                if translated:
                                    if screenstats:
                                        print('The RINEX 3 file has been downloaded. Try to make ', r2)
                                    fexists = g.new_rinex3_rinex2(rnx_filename,r2,dec_rate)
                                    #subprocess.call(['rm', '-f',rnx_filename]) # rnx
                        # this means the rinex 2 version exists
                        if fexists:
                             if screenstats:
                                 print('RINEX 2 created from v3', year, doy, ' Now remove RINEX 3 files and convert')
                             subprocess.call(['rm', '-f',rnx_filename]) # rnx
                             conv2snr(year, doy, station, isnr, orbtype,rate,dec_rate,archive,translator)
                        else:
                            print('Unsuccessful RINEX 3 retrieval/translation', year, doy)
                    else:
                        print(station, ' year:', year, ' doy:', doy, ' from: ', archive, ' rate:', rate, ' orb:', orbtype)
                        # this is rinex version 2 - finds rinex and converts it
                        conv2snr(year, doy, station, isnr, orbtype,rate,dec_rate,archive,translator)


def conv2snr(year, doy, station, option, orbtype,receiverrate,dec_rate,archive,translator):
    """
    convert RINEX files to SNR files

    2024 March 29: change location of logs directory to below REFL_CODE

    Parameters
    ----------
    year : int
        full year
    doy : int
        day of year
    option : int
        snr choice (66, 99 etc)
    orbtype : str
        orbit source (nav, gps, gnss, etc)
    receiverrate : int
        sampling interval of the GPS receiver, e.g. 1, 30, 15
    dec_rate : int
        decimation value to reduce file size
    archive : str
        external location (archive) of the rinex files
    translator : str
         hybrid, python, or fortran

    """
    xdir = os.environ['REFL_CODE']

    # universal location for the log directory
    logdir, logname = g.define_logdir(station,year,doy)

    exedir = os.environ['EXE'] # location of executables for fortran people

    screenstats = False # for now

    general_log = logdir + '/' + logname + '.gen'
    print('Minimal feedback is written to ', general_log)
    errorlog = logdir + '/' + logname

    log = open(general_log, 'w+')
    log.write("Receiver rate: {0:5s} \n".format(receiverrate))
    log.write("Decimation rate: {0:3.0f} \n".format(dec_rate))
    log.write("Archive: {0:10s} \n".format(archive))
    log.write("Orbits : {0:10s} \n".format(orbtype))
    log.write("Translator : {0:10s} \n".format(translator))
    csnr = str(option)
    # should be using a function for this
    cyyyy,cyy,cdoy = g.ydoych(year,doy)

    snrname_full =  quickname(station,year,cyy,cdoy,csnr)
    log.write("Creating : {0:s} \n".format(snrname_full))

    snrname_compressed = ''

    # if it exists, you should not have gotten to this code...
    snre = False
    #snrname_full, snrname_compressed, snre = g.define_and_xz_snr(station,year,doy,option)

    # use a local definition of fortran now that we are only using translator variable from
    # the functions that call conv2snr
    # 
    fortran = False
    if translator == 'fortran':
        fortran == True


    if (snre == True):
        log.write("The snrfile already exists: {0:50s} \n".format(snrname_full))
        print("The snrfile already exists: ", snrname_full)
    else:
        log.write("The snrfile does not exist: {0:50s} \n".format(snrname_full))
        d = g.doy2ymd(year,doy);
        month = d.month; day = d.day
        # new function to do the whole orbit thing
        foundit, f, orbdir, snrexe = g.get_orbits_setexe(year,month,day,orbtype,fortran)
        # if you have the orbit file, you can get the rinex file. First lets define the expected names
        print('Orbit file: ', orbdir + '/' + f)
        if foundit:
            # now you can look for a rinex file
            rinexfile,rinexfiled = g.rinex_name(station, year, month, day)
            # This goes to find the rinex file. I am changing it to allow
            # an archive preference
            if receiverrate == 'high':
                strip_snr = False # for now -
                #print('trying to find highrate file')
                file_name, foundit = k.rinex2_highrate(station, year, doy,archive,strip_snr)
            else:
                # added karnak librariies
                if (archive == 'all'):
                    foundrinex = False
                    for archivechoice in ['unavco','sopac','sonel']:
                        if (not foundrinex):
                            file_name,foundrinex = k.universal_rinex2(station, year, doy, archivechoice,screenstats)
                else:
                    file_name,foundrinex = k.universal_rinex2(station, year, doy, archive,screenstats)

                if foundrinex: #uncompress etc  to make o files ...
                    rinexfile, foundit2 = k.make_rinex2_ofiles(file_name) # translate

#           define booleans for various files
            oexist = os.path.isfile(orbdir + '/' + f) == True
            rexist = os.path.isfile(rinexfile) == True
            exc = exedir + '/teqc'
            texist = os.path.isfile(exc) == True
            if rexist:
                # decimate using teqc  if you have it
                if (texist) and (fortran) and (dec_rate > 0):
                    log.write("Decimating using teqc:  {0:3.0f}  seconds \n".format(dec_rate))
                    log.write('Unfortunately teqc removes Beidou data. Eventually I will remove this. \n')
                    rinexout = rinexfile + '.tmp'; cdec = str(dec_rate)
                    fout = open(rinexout,'w')
                    subprocess.call([exc, '-O.dec', cdec, rinexfile],stdout=fout)
                    fout.close() # needed?
                    status = subprocess.call(['mv','-f', rinexout, rinexfile])
            # if orbits and rinexfile exist
            if (oexist) and (rexist):
                snrname = g.snr_name(station, year,month,day,option)
                orbfile = orbdir + '/' + f
                # almost everyone uses hybrid. python only is heavily discouraged because
                # it is too slow. 
                if translator == 'hybrid':
                    g.make_snrdir(year,station) # make sure output directory exists
                    in1 = g.binary(rinexfile)
                    in2 = g.binary(snrname) # this file is made locally and moved later
                    in3 = g.binary(orbfile)
                    if (len(snrname) > 132) or (len(orbfile) > 132):
                        print('The orbit or SNR file name is too long.')
                        print('Make your environment variable names shorter.')
                        return
                    in4 = g.binary(str(option))
                    if (dec_rate > 0):
                        decr = str(dec_rate)
                    else:
                        decr = '0'
                    in5 = g.binary(decr) # decimation can be used in hybrid option
                    message = 'None '

                    in6 = g.binary(errorlog)
                    log.write('SNR file {0:50s} \n will use hybrid of python and fortran to make \n'.format( snrname))
                    # these are calls to the fortran codes that have been ported to be called from python
                    if (orbtype  == 'gps') or ('nav' in orbtype):
                        gpssnr.foo(in1,in2,in3,in4,in5,in6)
                    else:
                        if orbtype in ['ultra', 'wum', 'wum2']:
                            print('Using an ultrarapid orbit')
                            gnsssnrbigger.foo(in1,in2,in3,in4,in5,in6)
                        else:
                            gnsssnr.foo(in1,in2,in3,in4,in5,in6)
                else:
                    if (translator == 'fortran'):
                        t1=time.time()
                        try:
                            #subprocess.call([snrexe, rinexfile, snrname, orbfile, str(option)])
                            log.write('Using standalone fortran for translation  - separate log is used for stdout \n')
                            flogname = errorlog

                            flog = open(flogname, 'w+')
                            a=subprocess.run([snrexe, rinexfile, snrname, orbfile, str(option)],capture_output=True,text=True)
                            ddd = a.stdout; flog.write(ddd); flog.close()
                            status = subprocess.call(['rm','-f', rinexfile ])
                            status = subprocess.call(['xz', orbfile])
                        except:
                            log.write('Problem with making SNR file, check log {0:50s} \n'.format(flogname))
                        t2=time.time()
                    else:
                        log.write('SNR file {0:50s} \n will use python to make \n'.format( snrname))
                        log.write('Decimating will be done here instead of using teqc \n')
                        t1=time.time()
                        rnx2snr(rinexfile, orbfile,snrname,option,year,month,day,dec_rate,log)
                        t2=time.time()

                # remove the rinex file
                subprocess.call(['rm', '-f',rinexfile])

                if os.path.isfile(snrname):
#                make sure it exists and is non-zero size before moving it
                    if (os.stat(snrname).st_size == 0):
                        log.write('you created a zero file size which could mean a lot of things \n')
                        log.write('bad exe, bad snr option, do not really have the orbit file \n')
                        status = subprocess.call(['rm','-f', snrname ])
                    else:
                        log.write('A SNR file was created : {0:50s}  \n'.format(snrname_full))
                        print('\n')
                        print('SUCCESS: SNR file was created \n', snrname_full)
                        g.store_snrfile(snrname,year,station)
                        subprocess.call(['gzip', snrname_full])

                else:
                    # not sure why this is made here??? wasn't it created earlier?
                    print('No SNR file created - check ', errorlog, ' or ')
                    print( general_log , ' for why it failed.')
            else:
                print('Either the RINEX file or orbit file does not exist, so there is nothing to convert')
                log.write('Either the RINEX file or orbit file does not exist, so there is nothing to convert \n')
        else:
            print('The orbit file you requested does not exist.')

    log.close()

    return True

def satorb(week, sec_of_week, ephem):
    """
    Calculate GPS satellite orbits

    Parameters
    ----------
    week : integer
        GPS week

    sec_of_week : float
        GPS seconds of the week

    ephem : ephemeris block

    Returns
    -------
    numpy array
         the x,y,z, coordinates of the satellite in meters
         and relativity correction (also in meters), so you add, not subtract

    """

# redefine the ephem variable
    prn, week, Toc, Af0, Af1, Af2, IODE, Crs, delta_n, M0, Cuc,\
    ecc, Cus, sqrta, Toe, Cic, Loa, Cis, incl, Crc, perigee, radot, idot,\
    l2c, week, l2f, sigma, health, Tgd, IODC, Tob, interval = ephem
    sweek = sec_of_week
    # semi-major axis
    a = sqrta**2
    t = week*7*86400+sweek
    tk = t-Toe
    # no idea if Ryan Hardy is doing this correctly - it should be in a function
    tk  =  (tk - 302400) % (302400*2) - 302400
    n0 = np.sqrt(constants.mu/a**3)
    n = n0+ delta_n
    Mk = M0 + n*tk
    i = 0
    Ek = Mk
    E0 = Mk + ecc*np.sin(Mk)
    # solve kepler's equation
    while(i < 3 or np.abs(Ek-E0) > 1e-12):
        i +=1
        Ek = Mk + ecc*np.sin(E0)
        E0 = Mk + ecc*np.sin(Ek)
    nuk = np.arctan2(np.sqrt(1-ecc**2)*np.sin(Ek),np.cos(Ek)-ecc)
    Phik = nuk + perigee
    duk = Cus*np.sin(2*Phik)+Cuc*np.cos(2*Phik)
    drk = Crs*np.sin(2*Phik)+Crc*np.cos(2*Phik)
    dik = Cis*np.sin(2*Phik)+Cic*np.cos(2*Phik)
    uk = Phik + duk
    rk = a*(1-ecc*np.cos(Ek))+drk

    ik = incl+dik+idot*tk
    xkp = rk*np.cos(uk)
    ykp = rk*np.sin(uk)
    Omegak = Loa + (radot-constants.omegaEarth)*tk -constants.omegaEarth*Toe
    xk = xkp*np.cos(Omegak)-ykp*np.cos(ik)*np.sin(Omegak)
    yk = xkp*np.sin(Omegak)+ykp*np.cos(ik)*np.cos(Omegak)
    zk = ykp*np.sin(ik)
    # try this
    return np.array([xk, yk, zk])


def rnx2snr(obsfile, navfile,snrfile,snroption,year,month,day,dec_rate,log):
    """
    Converts a rinex v2.11 obs file using Joakim's rinex reading code

    Parameters
    ----------
    obsfile : str
        RINEX 2.11 filename
    navfile : str
        navigation file

    snrfile : str
        SNR filename

    snroption: integer
        kind of SNR file requested

    year : int
        full year
    month : int
        calendar month
    day : int
        calendar day
    dec_rate : int
        decimation rate in seconds

    """
    station = obsfile[0:4]
    #logname = 'logs/' + station + 'python.txt'
    #log = open(logname, 'w+')
    last3 = navfile[-3::]
    # figure out if you have a nav file or a sp3 file
    orbtype = 'sp3' # assume it is sp3
    if (last3 != 'SP3') and (last3 != 'sp3'):
        orbtype = 'nav'
    log.write("Orbit type {0:4s} \n".format(orbtype))
    log.write("File name {0:50s} \n".format(navfile))
    # these are the elevation angle limits I have been using for the various SNR formats
    emin,emax = elev_limits(snroption)

    exitQ = False
    obsdata, systemsatlists, prntoidx, obstypes, header, obstimes,gpstime = rinpy.processrinexfile(obsfile)
    obsdata = rinpy.separateobservables(obsdata, obstypes)
    obslist = obstypes['G'][:]
    # need to check to see what happens without coordinates
    key = 'APPROX POSITION XYZ'
    if key in header.keys():
        log.write('Cartesian coordinates are in the RINEX Header \n')
    else:
        log.write('RINEX file does not have station coordinates. Exiting \n')
        print('RINEX file does not have station coordinates. This is illegal. Exiting')
        return
    rv =  header['APPROX POSITION XYZ']
    recv = [float(i) for i in rv.split()]
    recv = np.array(recv)
    log.write("XYZ from header {0:15.5f} {1:15.5f} {2:15.5f} \n".format(recv[0],recv[1],recv[2]))
    if np.sum(np.abs(recv)) < 5:
        print('Your receiver coordinates are in the middle of the Earth. Exiting.')
        exitQ = True
        return

    lat, lon, h = g.xyz2llh(recv,1e-8) # returns lat/lon in radians
    up,East,North = g.up(lat,lon) # returns unit vector for UP


# set defaults
    s5exist = False; s1exist = False; s2exist = False;
    if 'S1' in obslist :
        s1exist = True
    if 'S2' in obslist :
        s2exist = True
    if 'S5' in obslist :
        s5exist = True
    if not s1exist and not s2exist:
        log.write('There are no S1 and no S2 data - this file is not useful for reflectometry \n')
        exitQ = True
    if (orbtype == 'nav'):
        gpssatlist = systemsatlists['G'][:]
        #print('GPS satellite list', gpssatlist)
        navorbits(navfile,obstimes,obsdata,obslist,prntoidx,gpssatlist,snrfile,s1exist,s2exist,s5exist,up,East,North,emin,emax,recv,dec_rate,log)
    else:
        log.write('Read the sp3 file \n'); sp3 = g.read_sp3file(navfile)
        _testing_sp3(gpstime,sp3,systemsatlists,obsdata,obstypes,prntoidx,year,month,day,emin,emax,snrfile,up,East,North,recv,dec_rate,log)
        #_test_sp3(gpstime,sp3,systemsatlists,obsdata,obstypes,prntoidx,year,month,day,emin,emax,snrfile,up,East,North,recv,dec_rate,log)

    #print('Closing python RINEX conversion log file:',logname)
    #log.close()

def navorbits(navfile,obstimes,observationdata,obslist,prntoidx,gpssatlist,snrfile,s1exist,s2exist,s5exist,up,East,North,emin,emax,recv,dec_rate,log):
    """
    Strandberg nav reading file?

    Parameters
    ---------
    navfile : string

    obstimes : ??

    observationdata :

    obslist :

    prn2oidx :

    gpssatlist :

    snrfile : str
        name of the output file

    s1exist :

    s2exist :

    s5exist :

    This is for GPS only files !
    navfile is nav broadcast ephemeris in RINEX format
    inputs are rinex info, obstimes, observationdata,prntoidx,gpssatlist
    various bits about SNR existence
    snrfile is output name
    log is for screen outputs - now going to a file
    """
    log.write('reading the ephemeris data \n')
    ephemdata = g.myreadnav(navfile)
    if len(ephemdata) == 0:
        log.write("Empty ephemeris or the file does not exist \n")
        return

    # change variable name to save typing
    a=obstimes
    if True:
        log.write('Opening output file for the SNR data \n')
        fout = open(snrfile, 'w+')
        K=len(obstimes)
        log.write('Number of epochs in the RINEX file {0:6.0f} \n '.format( K))
        log.write('Decimation rate {0:3.0f} \n'.format(dec_rate))

        with Bar('Processing RINEX', max=K,fill='@',suffix='%(percent)d%%') as bar:
            for i in range(0,K):
                bar.next()
                if np.remainder(i,200) == 0:
                    log.write('Epoch {0:6.0f} \n'.format( i))
            # sod is seconds of the day
                sod = 3600*a[i].hour + 60*a[i].minute + a[i].second
                if dec_rate > 0:
                    rem = sod % dec_rate
                else:
                    rem = 0
                if (rem == 0):
                    gweek, gpss = g.kgpsweek(a[i].year, a[i].month, a[i].day, a[i].hour, a[i].minute, a[i].second)
                    for sat in gpssatlist:
                        s1,s2,s5 = readSNRval(s1exist,s2exist,s5exist,observationdata,prntoidx,sat,i)
                        if (s1 > 0):
                            closest = g.myfindephem(gweek, gpss, ephemdata, sat)
                            if len(closest) > 0:
                                satv = satorb_prop(gweek, gpss, sat, recv, closest)
                                r=np.subtract(satv,recv) # satellite minus receiver vector
                                eleA = g.elev_angle(up, r)*180/np.pi
                                azimA = g.azimuth_angle(r, East, North)
                                if (eleA >= emin) and (eleA <= emax):
                                    fout.write("{0:3.0f} {1:10.4f} {2:10.4f} {3:10.0f} {4:7.2f} {5:7.2f} {6:7.2f} {7:7.2f} {8:7.2f} \n".format(sat,eleA, azimA, sod,0, 0, s1,s2, s5))
        fout.close()
    else:
        log.write('There was some kind of problem with your file, exiting ...\n')
        print('There was some kind of problem with your file, exiting ...')

def readSNRval(s1exist,s2exist,s5exist,observationdata,prntoidx,sat,i):
    """
    what it looks like only reads GPS data for now
    interface between Joakim's code and mine ...

    Parameters
    ----------
    s1exist : boolean

    s2exist : boolean

    s5exist : boolean

    Returns
    -------
    s1 :
    s2 :
    s5 :

    """
    s1 = 0; s2 = 0; s5 = 0 # set defaults
    if s1exist:
        s1 = observationdata['G']['S1'][i, prntoidx['G'][sat]]
        if np.isnan(s1):
           s1 = 0
    if s2exist:
        s2 = observationdata['G']['S2'][i, prntoidx['G'][sat]]
        if np.isnan(s2):
            s2 = 0
    if s5exist:
        s5 = observationdata['G']['S5'][i, prntoidx['G'][sat]]
        if np.isnan(s5):
            s5 = 0
    return s1, s2, s5


def satorb_prop(week, secweek, prn, rrec0, closest_ephem):
    """
    Calculates and returns geometric range (in metres) given
    time (week and sec of week), prn, receiver coordinates (cartesian, meters)
    this assumes someone was nice enough to send you the closest ephemeris
    returns the satellite coordinates as well, so you can use htem
    in the A matrix

    Parameters
    ----------
    week : integer
        GPS week

    secweek : integer
        GPS second of the week

    prn : integer
        satellite number

    rrec0 : 3vector
        receiver coordinates, meters

    Returns
    -------
    SatOrbn : 3vector
        floats, Cartesian location of satellite in meters [x,y,z]

    """
    error = 1

    # might as well start with 70 milliseconds
    SatOrb = satorb(week, secweek-0.07, closest_ephem)
    # first estimate of the geometric range
    geo= g.norm(SatOrb-rrec0)

    deltaT = g.norm(SatOrb - rrec0)/constants.c
    k=0
    #while (error > 1e-8) or (k < 2):
    # should not need more than two iterations, since i am
    #starting with 70 msec
    while (k < 2):
        SatOrb = satorb(week, secweek-deltaT, closest_ephem)
        Th = -constants.omegaEarth * deltaT
        xs = SatOrb[0]*np.cos(Th)-SatOrb[1]*np.sin(Th)
        ys = SatOrb[0]*np.sin(Th)+SatOrb[1]*np.cos(Th)
        SatOrbn = [xs, ys, SatOrb[2]]
        # try this ???
        geo = g.norm(SatOrbn-rrec0)
        deltaT_new = g.norm(SatOrbn-rrec0)/constants.c
        error = np.abs(deltaT - deltaT_new)
        deltaT = deltaT_new
        k += 1
    return SatOrbn


def satorb_prop_sp3(iX,iY,iZ,recv,Tp,ij):
    """
    for satellite number prn
    and receiver coordinates rrec0
    find the x,y,z coordinates at time secweek

    Parameters
    ----------
    iX : float

    iY : float

    iZ : float
    recv : 3 vector, float

    Tp :

    ij :

    sp3 has the orbit information in it
    """
    # start wit 70 milliseconds as the guess for the transmission time
    nx = iX(Tp[ij]-0.07); ny = iY(Tp[ij]-0.07); nz = iZ(Tp[ij]-0.07)
    oE = constants.omegaEarth
    c = constants.c
    # get initial deltaA
    SatOrb=np.array([nx,ny,nz]).T
    r=np.subtract(SatOrb,recv)
    tau = g.norm(r)/c

    error = 0
    k = 0
    while (k < 2):
        nx = iX(Tp[ij]-tau); ny = iY(Tp[ij]-tau); nz = iZ(Tp[ij]-tau)
        SatOrb=np.array([nx,ny,nz]).T
        Th = -oE * tau
        xs = SatOrb[0]*np.cos(Th)-SatOrb[1]*np.sin(Th)
        ys = SatOrb[0]*np.sin(Th)+SatOrb[1]*np.cos(Th)
        SatOrbn = np.array([xs, ys, SatOrb[2]]).T
        tau=g.norm(SatOrbn-recv)/c
        k+=1

    return SatOrbn

def _test_sp3(gpstime,sp3,systemsatlists,obsdata,obstypes,prntoidx,year,month,day, emin,emax,outputfile,up,East,North,recv,dec_rate,log):
    """
    inputs are gpstime( numpy array with week and sow)
    sp3 is what has been read from the sp3 file
    columsn are satNu, week, sow, x, y, z (in meters)
    log is for comments
    """
    checkD = False
    if dec_rate > 0:
        checkD = True
        log.write('You are decimating \n')
    # epoch at the beginning of the day of your RINEX file
    gweek0, gpssec0 = g.kgpsweek(year, month,day,0,0,0 )

    ll = 'quadratic'
#   will store in this variable, then sort it before writing out to a file
    saveit = np.empty(shape=[0,11] )
    fout = open(outputfile, 'w+')
    for con in ['G','E','R','C']:
        if con in obstypes:
            log.write('Good news - found data for constellation {0:s} \n'.format( con))
            obslist = obstypes[con][:]
            satlist = systemsatlists[con][:]
            #print(satlist)
            for prn in satlist:
                addon = g.findConstell(con) # 100,200,or 300 for R,E, and C
                log.write('Constellation {0:1s} Satellite {1:2.0f}  Addon {2:3.0f} \n'.format( con, prn, addon))
                # window out the data for this satellite
                m = sp3[:,0] == prn + addon
                x = sp3[m,3]
                if len(x) > 0:
                    sp3_week = sp3[m,1] ; sp3_sec = sp3[m,2]
                    x = sp3[m,3] ; y = sp3[m,4] ; z = sp3[m,5]
                # fit the orbits for this satellite
                    t=sp3_sec
                    iX= interp1d(t, x, ll,bounds_error=False,fill_value='extrapolate')
                    iY= interp1d(t, y, ll,bounds_error=False,fill_value='extrapolate')
                    iZ= interp1d(t, z, ll,bounds_error=False,fill_value='extrapolate')
        # get the S1 data for this satellite
                    if 'S1' in obslist:
                        s1 = obsdata[con]['S1'][:, prntoidx[con][prn]]

        # indices when there are no data for this satellite
                    ij = np.isnan(s1)
        # indices when there are data in the RINEX file - this way you do not compute
        # orbits unless there are data.
                    not_ij = np.logical_not(ij)
                    Tp = gpstime[not_ij,1] # only use the seconds of the week for now
                    s1 = s1[not_ij];
                    #print(s1.shape)
                    emp = np.zeros(shape=[len(s1),1],dtype=float)
        # get the rest of the SNR data in a function
                    s2,s5,s6,s7,s8 = extract_snr(prn, con, obslist,obsdata,prntoidx,not_ij,emp)

        # make sure there are no nan values in s2 or s5

                    nepochs = len(Tp)
                    for ij in range(0,nepochs):
                        TT = 0 # default value
                        if checkD:
                            TT = Tp[ij]  % dec_rate # get the modulus
                        if TT == 0:
                            SatOrb = satorb_prop_sp3(iX,iY,iZ,recv,Tp,ij)
                            r=np.subtract(SatOrb,recv)
                            azimA = g.azimuth_angle(r, East, North)
                            eleA = g.elev_angle(up, r)*180/np.pi
                            # 2021 october 26
                            # thank you to andrea gatti for pointing out the mistake
                            if (eleA >= emin) and (eleA <= emax):
                                fout.write("{0:3.0f} {1:10.4f} {2:10.4f} {3:10.0f} {4:7.2f} {5:7.2f} {6:7.2f} {7:7.2f} {8:7.2f} {9:7.2f} {10:7.2f} \n".format(
                                    prn+addon,eleA,azimA,Tp[ij]-gpssec0, 0,float(s6[ij]),s1[ij],float(s2[ij]),float(s5[ij]),float(s7[ij]),float(s8[ij]) ))
                                #fout.write("{0:3.0f} {1:10.4f} {2:10.4f} {3:10.0f} {4:7.2f} {5:7.2f} {6:7.2f} {7:7.2f} {8:7.2f} {9:7.2f} {10:7.2f} \n".format(
                                #    prn+addon,eleA,azimA,Tp[ij]-gpssec0, 0,float(s6[ij]),s1[ij],float(s2[ij]),float(s5[ij]),float(s6[ij]),float(s7[ij]) ))
                else:
                    log.write('This satellite is not in the orbit file. {0:3.0f} \n'.format(prn))
        else:
            log.write('No data for constellation {0:1s} \n'.format(con))
    # print('sort by time')
    # tried saving to variable but it was very slow
    #ne = np.array([prn,eleA,azimA,Tp[ij],0,0,s1[ij],s2[ij],s5[ij],0,0])
    #saveit = np.vstack((saveit,ne))
    #i = np.argsort(saveit[:,3])
    # apply that sort to variable with shorter name
    #s = saveit[i,:]
    log.write('write SNR data to file \n')
    fout.close()


def extract_snr(prn, con, obslist,obsdata,prntoidx,not_ij,emp):
    """
    """
    # defaults are zero arrays
    s2 = emp; s5 = emp; s6 = emp; s7 = emp; s8 = emp
    if 'S2' in obslist:
        s2 = obsdata[con]['S2'][:, prntoidx[con][prn]]
        s2 = s2[not_ij] ; is2 = np.isnan(s2); s2[is2] = 0
    if 'S5' in obslist:
        s5 = obsdata[con]['S5'][:, prntoidx[con][prn]]
        s5 = s5[not_ij]
        is5 = np.isnan(s5); s5[is5] = 0
    if 'S6' in obslist:
        s6 = obsdata[con]['S6'][:, prntoidx[con][prn]]
        s6 = s6[not_ij]
        is6 = np.isnan(s6); s6[is6] = 0
    if 'S7' in obslist:
        s7 = obsdata[con]['S7'][:, prntoidx[con][prn]]
        s7 = s7[not_ij]
        is7 = np.isnan(s7); s7[is7] = 0
    if 'S8' in obslist:
        s8 = obsdata[con]['S8'][:, prntoidx[con][prn]]
        s8 = s8[not_ij]

    return s2,s5,s6,s7,s8

def elev_limits(snroption):
    """
    For given SNR option, returns elevation angle limits

    Parameters
    ------------
    snroption : integer
        snr file delimeter

    Returns
    ----------
    emin: float
        minimum elevation angle (degrees)
    emax: float
        maximum elevation angle (degrees)

    """

    if (snroption == 99):
        emin = 5; emax = 30
    elif (snroption == 50):
        emin = 0; emax = 10
    elif (snroption == 66):
        emin = 0; emax = 30
    elif (snroption == 88):
        emin = 5; emax = 90
    else:
        emin = 5; emax = 30

    return emin, emax


def _testing_sp3(gpstime,sp3,systemsatlists,obsdata,obstypes,prntoidx,year,month,day, emin,emax,outputfile,up,East,North,recv,dec_rate,log):
    """
    inputs are gpstime( numpy array with week and sow)
    sp3 is what has been read from the sp3 file
    columsn are satNu, week, sow, x, y, z (in meters)
    log is for comments
    """
    checkD = False
    if dec_rate > 0:
        checkD = True
        log.write('You are decimating \n')
    # epoch at the beginning of the day of your RINEX file
    gweek0, gpssec0 = g.kgpsweek(year, month,day,0,0,0 )

    ll = 'quadratic'
#   will store in this variable, then sort it before writing out to a file
    saveit = np.empty(shape=[0,11] )
    fout = open(outputfile, 'w+')
    NsatT = 0
    # make a dictionary for constellation name
    sname ={}; sname['G']='GPS' ; sname['R'] = 'GLONASS'; sname['E'] = 'GALILEO'; sname['C']='BEIDOU'
    for con in ['G','E','R','C']:
        if con in obstypes:
            satL = len(systemsatlists[con][:])
            satS = 'Processing ' + sname[con]
            with Bar(satS, max=satL,fill='@',suffix='%(percent)d%%') as bar:
                log.write('Good news - found data for constellation {0:s} \n'.format( con))
                obslist = obstypes[con][:]
                satlist = systemsatlists[con][:]
                for prn in satlist:
                    bar.next()
                    addon = g.findConstell(con) # 100,200,or 300 for R,E, and C
                    log.write('Constellation {0:1s} Satellite {1:2.0f}  Addon {2:3.0f} \n'.format( con, prn, addon))
                # window out the data for this satellite
                    m = sp3[:,0] == prn + addon
                    x = sp3[m,3]
                    if len(x) > 0:
                        sp3_week = sp3[m,1] ; sp3_sec = sp3[m,2]
                        x = sp3[m,3] ; y = sp3[m,4] ; z = sp3[m,5]
                # fit the orbits for this satellite
                        t=sp3_sec
                        iX= interp1d(t, x, ll,bounds_error=False,fill_value='extrapolate')
                        iY= interp1d(t, y, ll,bounds_error=False,fill_value='extrapolate')
                        iZ= interp1d(t, z, ll,bounds_error=False,fill_value='extrapolate')
        # get the S1 data for this satellite
                        if 'S1' in obslist:
                            s1 = obsdata[con]['S1'][:, prntoidx[con][prn]]

        # indices when there are no data for this satellite
                        ij = np.isnan(s1)
        # indices when there are data in the RINEX file - this way you do not compute
        # orbits unless there are data.
                        not_ij = np.logical_not(ij)
                        Tp = gpstime[not_ij,1] # only use the seconds of the week for now
                        s1 = s1[not_ij];
                    #print(s1.shape)
                        emp = np.zeros(shape=[len(s1),1],dtype=float)
        # get the rest of the SNR data in a function
                        s2,s5,s6,s7,s8 = extract_snr(prn, con, obslist,obsdata,prntoidx,not_ij,emp)
        # make sure there are no nan values in s2 or s5

                        nepochs = len(Tp)
                        for ij in range(0,nepochs):
                            TT = 0 # default value
                            if checkD:
                                TT = Tp[ij]  % dec_rate # get the modulus
                            if TT == 0:
                                SatOrb = satorb_prop_sp3(iX,iY,iZ,recv,Tp,ij)
                                r=np.subtract(SatOrb,recv)
                                azimA = g.azimuth_angle(r, East, North)
                                eleA = g.elev_angle(up, r)*180/np.pi
                                if (eleA >= emin) and (eleA <= emax):
                                    # bug reported by Andrea Gatti. 2021 October 26
                                    fout.write("{0:3.0f} {1:10.4f} {2:10.4f} {3:10.0f} {4:7.2f} {5:7.2f} {6:7.2f} {7:7.2f} {8:7.2f} {9:7.2f} {10:7.2f} \n".format(
                                        prn+addon,eleA,azimA,Tp[ij]-gpssec0, 0,float(s6[ij]),s1[ij],float(s2[ij]),float(s5[ij]),float(s7[ij]),float(s8[ij]) ))
                    else:
                        log.write('This satellite is not in the orbit file. {0:3.0f} \n'.format(prn))
        else:
            log.write('No data for constellation {0:1s} \n'.format(con))
    log.write('write SNR data to file \n')
    fout.close()


def the_makan_option(station,cyyyy,cyy,cdoy):
    """
    this ugly looking code checks a bazillion versions of RINEX versions
    (Z, gz, regular, hatanaka) both in the working directory and in an external rinex area
    $REFL_CODE/rinex/station/year

    turns whatever it finds into a regular RINEX file in the working directory
    that file WILL be deleted, but it will not delete those stored externally.

    Parameters
    ----------
    station : str
        station name (4 ch)
    cyyyy : str
        4 ch year
    cyy : str
        two ch year
    cdoy : str
        three ch day of year
    """
    missing = True
    crnxpath = g.hatanaka_version()  # where hatanaka will be
    r = station + cdoy + '0.' + cyy + 'o'
    rd = station + cdoy + '0.' + cyy + 'd'
    print(r,rd)

    locdir=  os.environ['REFL_CODE'] + '/rinex/' + station + '/' + cyyyy + '/'
    locdir2= os.environ['REFL_CODE'] + '/' + cyyyy + '/rinex/' + station + '/'
    #
    #locdir2= os.environ['RINEX'] + station + '/' + cyyyy + '/'
    #locdir3= os.environ['RINEX'] + station.upper() + '/' + cyyyy + '/'

    print('Will look for files in the working directory and ', locdir )
    # I was testing this ... but have not finished
    #so_many_permutations(r,rd,locdir, crnxpath)

    if os.path.exists(r):
        #print('type 1')
        missing = False

    if os.path.exists(r + '.gz') and missing:
        #print('type 2')
        subprocess.call(['gunzip', r + '.gz'])
        missing = False

    if os.path.exists(r + '.Z') and missing:
        #print('type 3')
        subprocess.call(['uncompress', r + '.Z'])
        missing = False

    if os.path.exists(rd) and missing:
        #print('type 4')
        if os.path.exists(crnxpath):
            subprocess.call([crnxpath,rd])
            subprocess.call(['rm',rd])
            missing = False
        else:
            g.hatanaka_warning(); return

    if os.path.exists(rd + '.gz') and missing:
        #print('type 5')
        subprocess.call(['gunzip', rd + '.gz'])
        if os.path.exists(crnxpath):
            subprocess.call([crnxpath,rd])
            subprocess.call(['rm',rd])
            missing = False
        else:
            g.hatanaka_warning();

    if os.path.exists(rd + '.Z') and missing:
        #print('type 6')
        subprocess.call(['uncompress', rd + '.Z'])
        if os.path.exists(crnxpath):
            subprocess.call([crnxpath,rd])
            subprocess.call(['rm',rd])
            missing = False
        else:
            g.hatanaka_warning()

    if os.path.exists(locdir + r) and missing:
        #print('type 7')
        subprocess.call(['cp', '-f',locdir + r,'.'])
        missing = False

    if os.path.exists(locdir + r + '.gz') and missing:
        #print('type 8')
        subprocess.call(['cp', '-f',locdir + r + '.gz' ,'.'])
        subprocess.call(['gunzip', r + '.gz'])
        missing = False

    if os.path.exists(locdir + r + '.Z') and missing:
        subprocess.call(['cp', '-f',locdir + r + '.Z','.'])
        subprocess.call(['uncompress', r + '.Z'])
        missing = False

    if os.path.exists(locdir + rd) and missing:
        subprocess.call(['cp','-f',locdir + rd,'.'])
        if os.path.exists(crnxpath):
            subprocess.call([crnxpath,rd])
            subprocess.call(['rm',rd])
            missing = False
        else:
            g.hatanaka_warning();

    if os.path.exists(locdir + rd + '.Z') and missing:
        print('here?')
        subprocess.call(['cp','-f',locdir + rd + '.Z','.'])
        subprocess.call(['uncompress', rd + '.Z'])
        if os.path.exists(crnxpath):
            subprocess.call([crnxpath,rd])
            subprocess.call(['rm',rd])
            missing = False
        else:
            g.hatanaka_warning();

    if os.path.exists(locdir + rd + '.gz') and missing:
        subprocess.call(['cp','-f',locdir + rd + '.gz','.'])
        subprocess.call(['gunzip',rd + '.gz'])
        if os.path.exists(crnxpath):
            subprocess.call([crnxpath,rd])
            subprocess.call(['rm',rd])
            missing = False
        else:
            g.hatanaka_warning()

def go_from_crxgz_to_rnx(c3gz,deletecrx=True):
    """
    checks to see if rinex3 file exists, gunzip if necessary,
    run hatanaka, if necessary

    Parameters
    ----------
    c3gz : str
        filename for a gzipped RINEX 3 Hatanaka file

    deletecrx = bool
        whether to delete the crx file

    Returns
    -------
    translated : bool
        if file successfully found and available

    rnx : str
        name of gunzipped and decompressed RINEX 3

    """
    translated = False # assume failure
    c3 = c3gz[:-3] # crx filename
    rnx = c3.replace('crx','rnx') # rnx filename
    # gunzip
    if os.path.exists(c3gz):
        subprocess.call(['gunzip', c3gz])

    # executable
    crnxpath = g.hatanaka_version()
    if not os.path.exists(crnxpath):
        g.hatanaka_warning()
    else:
        if os.path.exists(c3): # file exists
            subprocess.call([crnxpath,c3])
    if os.path.exists(rnx): # file exists
        translated = True
        if deletecrx:
            #print('remove Hatanaka compressed file')
            subprocess.call(['rm','-f',c3])

    return translated, rnx

def get_local_rinexfile(rfile,localpath2):
    """
    look for a plain or gzipped version of the RINEX 2.11 file in the year subdirectories
    copies it to the local directory.  this method stops the code from deleting your rinex 
    files. As of 2023 September 19, it should also look for Hatanaka files.

    localpath2 =  os.environ['REFL_CODE'] + '/' + cyyyy + '/rinex/' + station + '/'

    This is unlikely to work for uppercase RINEX files. Try the mk option

    Parameters
    ----------
    rfile : str
        version2 rinexfile name

    localpath2 : str
        another location of the file (meant to be as defined above)

    Returns
    -------
    allgood : bool
        whether file found
    """
    crnxpath = g.hatanaka_version()
    allgood = False
    # hatanaka filename has d at end instead of o
    rd = rfile[0:-1] + 'd'

    # look for gzip version in local directory first
    if os.path.exists(rfile + '.gz'):
        print('found ', rfile + '.gz')
        subprocess.call(['gunzip', rfile + '.gz'])
        allgood = True


    # then look for unix compressed version local directory 
    if os.path.exists(rfile + '.Z') and not allgood:
        print('found ', rfile + '.Z')
        subprocess.call(['uncompress', rfile + '.Z'])
        allgood = True 

    # then look for hatanaka compressed version local directory 
    if os.path.exists(rd) and not allgood:
        print('found ', rd)
        if os.path.exists(crnxpath):
            subprocess.call([crnxpath,rd])
            if os.path.exists(rfile):
                allgood = True
                subprocess.call(['rm',rd])

    # then look for hatanaka compressed and unix compressed version local directory 
    if os.path.exists(rd + '.Z') and not allgood:
        print('found ', rd + '.Z')
        if os.path.exists(crnxpath):
            subprocess.call(['uncompress', rd + '.Z'])
            subprocess.call([crnxpath,rd])
            if os.path.exists(rfile):
                allgood = True
                subprocess.call(['rm',rd])

    # now check in 
    if not allgood:
        print('Checking for the file in : ', localpath2)
        r = localpath2 + rfile
        # hatanaka version in REFL_CODE
        rdd = localpath2 + rd

        if os.path.exists(r):
            allgood = True
        # cp to the local directory
            print('copying RINEX file to local area')
            subprocess.call(['cp',r,'.'])
        else:
       # did not find normal rinex, so look for gzip version
            if os.path.exists(r + '.gz'):
                subprocess.call(['gunzip', r + '.gz'])
                if os.path.exists(r):
                    print('copying RINEX file to local area')
                    subprocess.call(['cp',r,'.'])
                    # now gzip the original file ...
                    subprocess.call(['gzip', r ])
                    allgood = True
            # if hatanaka is stored locally
            if os.path.exists(rdd) and not allgood:
                # copy from REFL_CODE to local dir
                # and convert
                print('copying Hatanaka file to local area')
                subprocess.call(['cp',rdd,'.'])
                subprocess.call([crnxpath,rd])
                if os.path.exists(rfile):
                    allgood = True
                    subprocess.call(['rm',rd])

    return allgood

def print_archives():
    """
    feeble attempt to print list of archives to screen ...

    """
    archive_list_rinex3 = ['unavco', 'cddis', 'bev', 'bkg', 'ga', 'epn', 'bfg','sonel','all','unavco2','nrcan','gfz','ignes']
    archive_list = ['sopac', 'unavco', 'sonel',  'nz', 'ga', 'bkg', 'jeff',
                    'ngs', 'nrcan', 'special', 'bev', 'jp', 'all','unavco2','cddis']
    w = 'RINEX 3 archives '
    for i in range(0,len(archive_list_rinex3)):
        w = w + archive_list_rinex3[i] + ' '
    w = w + '\n RINEX 2.11 archives '
    for i in range(0,len(archive_list)):
        w = w + archive_list[i] + ' '
    msg = w
    return msg



#from .gnssrefl0.gps import *
#from .gnssrefl0.read_snr_files.py import *
#from .gnssrefl0.rinpy import *
#from .gnssir_guts import *
#from .refraction import *
#from .rnx2snr import *
#from .quick_read_snr import *

